from typing import List

import math
import numpy as np

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.manager.traffic_manager import PGTrafficManager
from metadrive.utils import get_np_random, random_string
from metadrive.utils.utils import deprecation_warning


class Vehicle:
    """
    A moving vehicle on a road, and its kinematics.

    The vehicle is represented by a dynamical system: a modified bicycle model.
    It's state is propagated depending on its steering and acceleration actions.
    """

    COLLISIONS_ENABLED = True
    """ Enable collision detection between vehicles """

    LENGTH = 5.0
    """ Vehicle length [m] """
    WIDTH = 2.0
    """ Vehicle width [m] """
    DEFAULT_SPEEDS = [23, 25]
    """ Range for random initial speeds [m/s] """
    MAX_SPEED = 40.
    """ Maximum reachable speed [m/s] """
    def __init__(
        self,
        traffic_mgr: PGTrafficManager,
        position: List,
        heading: float = 0,
        speed: float = 0,
        np_random: np.random.RandomState = None,
        name: str = None
    ):
        raise DeprecationWarning("We don't use kinematics from highway now")
        deprecation_warning("Vehicle", "Policy Class", error=False)

        self.name = random_string() if name is None else name
        self.traffic_mgr = traffic_mgr
        self._position = np.array(position).astype('float')
        self.heading = heading
        self.speed = speed
        # self.lane_index, _ = self.traffic_mgr.current_map.road_network.get_closest_lane_index(
        #     self.position
        # ) if self.traffic_mgr else (np.nan, np.nan)
        # self.lane = self.traffic_mgr.current_map.road_network.get_lane(self.lane_index) if self.traffic_mgr else None
        self.action = {'steering': 0, 'acceleration': 0}
        self.crashed = False
        # self.log = []
        # self.history = deque(maxlen=30)
        self.np_random = np_random if np_random else get_np_random()

    # def update_lane_index(self, lane_index, lane):
    #     raise ValueError("Deprecated!")
    #     self.lane_index = lane_index
    #     self.lane = lane

    @property
    def position(self):
        if self._position is None:
            return np.array([np.nan, np.nan])
        else:
            return self._position.copy()

    def set_position(self, pos):
        self._position = np.asarray(pos).copy()

    # @classmethod
    # def make_on_lane(cls, traffic_manager: TrafficManager, lane_index: LaneIndex, longitudinal: float, speed: float = 0):
    #     """
    #     Create a vehicle on a given lane at a longitudinal position.
    #
    #     :param traffic_manager: the road where the vehicle is driving
    #     :param lane_index: index of the lane where the vehicle is located
    #     :param longitudinal: longitudinal position along the lane
    #     :param speed: initial speed in [m/s]
    #     :return: A vehicle with at the specified position
    #     """
    #     lane = traffic_manager.current_map.road_network.get_lane(lane_index)
    #     if speed is None:
    #         speed = lane.speed_limit
    #     return cls(traffic_manager, lane.position(longitudinal, 0), lane.heading_at(longitudinal), speed)

    @classmethod
    def create_random(
        cls,
        traffic_mgr: PGTrafficManager,
        lane: AbstractLane,
        longitude: float,
        speed: float = None,
        random_seed=None
    ):
        """
        Create a random vehicle on the road.

        The lane and /or speed are chosen randomly, while longitudinal position is chosen behind the last
        vehicle in the road with density based on the number of lanes.

        :param longitude: the longitude on lane
        :param lane: the lane where the vehicle is spawn
        :param traffic_mgr: the traffic_manager where the vehicle is driving
        :param speed: initial speed in [m/s]. If None, will be chosen randomly
        :return: A vehicle with random position and/or speed
        """
        if speed is None:
            speed = traffic_mgr.np_random.uniform(Vehicle.DEFAULT_SPEEDS[0], Vehicle.DEFAULT_SPEEDS[1])
        v = cls(
            traffic_mgr,
            list(lane.position(longitude, 0)),
            lane.heading_theta_at(longitude),
            speed,
            # random_seed=get_np_random(random_seed)
            # np_random=get_np_random(random_seed)
        )
        return v

    @classmethod
    def create_from(cls, vehicle: "Vehicle") -> "Vehicle":
        """
        Create a new vehicle from an existing one.

        Only the vehicle dynamics are copied, other properties are default.

        :param vehicle: a vehicle
        :return: a new vehicle at the same dynamical state
        """
        v = cls(vehicle.traffic_mgr, vehicle.position, vehicle.heading, vehicle.speed)
        return v

    # def act(self, action: Union[dict, str] = None) -> None:
    #     """
    #     Store an action to be repeated.

    def step(self, dt: float, action) -> None:
        """
        Propagate the vehicle state given its actions.

        Integrate a modified bicycle model with a 1st-order response on the steering wheel dynamics.
        If the vehicle is crashed, the actions are overridden with erratic steering and braking until complete stop.
        The vehicle's current lane is updated.

        :param dt: timestep of integration of the model [s]
        """
        assert isinstance(action, dict)
        # self.action = action
        action = self.clip_actions(action)
        delta_f = action['steering']
        beta = np.arctan(1 / 2 * np.tan(delta_f))
        v = self.speed * np.array([math.cos(self.heading + beta), math.sin(self.heading + beta)])

        self._position += v * dt

        self.heading += self.speed * math.sin(beta) / (self.LENGTH / 2) * dt
        self.speed += action['acceleration'] * dt
        # for performance reason,
        # self.on_state_update()

    def clip_actions(self, action) -> None:
        if self.crashed:
            action['steering'] = 0
            action['acceleration'] = -1.0 * self.speed
        action['steering'] = float(action['steering'])
        action['acceleration'] = float(action['acceleration'])
        if self.speed > self.MAX_SPEED:
            action['acceleration'] = min(action['acceleration'], 1.0 * (self.MAX_SPEED - self.speed))
        elif self.speed < -self.MAX_SPEED:
            action['acceleration'] = max(action['acceleration'], 1.0 * (self.MAX_SPEED - self.speed))
        return action

    @property
    def direction(self) -> np.ndarray:
        return np.array([math.cos(self.heading), math.sin(self.heading)])

    @property
    def velocity(self) -> np.ndarray:
        return self.speed * self.direction

    @property
    def heading_theta(self):
        return self.heading

    # def on_state_update(self) -> None:
    #     new_l_index, _ = self.traffic_manager.current_map.road_network.get_closest_lane_index(self.position)
    #     self.lane_index = new_l_index
    #     self.lane = self.traffic_manager.current_map.road_network.get_lane(self.lane_index)

    # def lane_distance_to(self, vehicle: "Vehicle", lane: AbstractLane = None) -> float:
    #     """
    #     Compute the signed distance to another vehicle along a lane.
    #
    #     :param vehicle: the other vehicle
    #     :param lane: a lane
    #     :return: the distance to the other vehicle [m]
    #     """
    #     deprecation_warning(1, 2, True)
    #     if not vehicle:
    #         return np.nan
    #     if not lane:
    #         lane = self.lane
    #     return lane.local_coordinates(vehicle.position)[0] - lane.local_coordinates(self.position)[0]

    # def check_collision(self, other: Union['Vehicle', 'TrafficSign']) -> None:
    #     """
    #     Check for collision with another vehicle.
    #
    #     :param other: the other vehicle or object
    #     """
    #     if self.crashed or other is self:
    #         return
    #
    #     if isinstance(other, Vehicle):
    #         if not self.COLLISIONS_ENABLED or not other.COLLISIONS_ENABLED:
    #             return
    #
    #         if self._is_colliding(other):
    #             self.speed = other.speed = min([self.speed, other.speed], key=abs)
    #             self.crashed = other.crashed = True
    #     elif isinstance(other, TrafficSign):
    #         if not self.COLLISIONS_ENABLED:
    #             return
    #
    #         if self._is_colliding(other):
    #             self.speed = min([self.speed, 0], key=abs)
    #             self.crashed = other.hit = True
    #     elif isinstance(other, TrafficSign):
    #         if self._is_colliding(other):
    #             other.hit = True

    # def _is_colliding(self, other):
    #     # Fast spherical pre-check
    #     if distance_greater(other.position, self.position, self.LENGTH):
    #         return False
    #     # Accurate rectangular check
    #     return utils.rotated_rectangles_intersect(
    #         (self.position, 0.9 * self.LENGTH, 0.9 * self.WIDTH, self.heading),
    #         (other.position, 0.9 * other.LENGTH, 0.9 * other.WIDTH, other.heading)
    #     )

    # @property
    def destination(self) -> np.ndarray:
        if getattr(self, "route", None):
            last_lane = self.traffic_mgr.current_map.road_network.get_lane(self.route[-1])
            return last_lane.position(last_lane.length, 0)
        else:
            return self.position

    # #
    # @property
    # def destination_direction(self) -> np.ndarray:
    #     if (self.destination != self.position).any():
    #         return (self.destination - self.position) / norm(*(self.destination - self.position))
    #     else:
    #         return np.zeros((2, ))

    # @property
    # def on_road(self) -> bool:
    #     """ Is the vehicle on its current lane, or off-traffic_manager ? """
    #     return self.lane.on_lane(self.position)
    #
    # def front_distance_to(self, other: "Vehicle") -> float:
    #     return self.direction.dot(other.position - self.position)

    def to_dict(self, origin_vehicle: "Vehicle" = None, observe_intentions: bool = True) -> dict:
        d = {
            'presence': 1,
            'x': self.position[0],
            'y': self.position[1],
            'vx': self.velocity[0],
            'vy': self.velocity[1],
            'cos_h': self.direction[0],
            'sin_h': self.direction[1],
            # 'cos_d': self.destination_direction[0],
            # 'sin_d': self.destination_direction[1]
        }
        if not observe_intentions:
            d["cos_d"] = d["sin_d"] = 0
        if origin_vehicle:
            origin_dict = origin_vehicle.to_dict()
            for key in ['x', 'y', 'vx', 'vy']:
                d[key] -= origin_dict[key]
        return d

    def destroy(self):
        self.traffic_mgr = None
        self._position = None
        self.heading = None
        self.speed = None
        self.lane_index = None
        self.lane = None
        self.action = None
        self.crashed = False
        self.log = None
        self.history = None
        self.np_random = None

    def __str__(self):
        return "{} #{}: {}".format(self.__class__.__name__, id(self) % 1000, self.position)

    def __repr__(self):
        return self.__str__()



import copy
import datetime
import logging
import os
import sys
import time

import numpy as np
from panda3d.bullet import BulletBodyNode

from metadrive.constants import MetaDriveType


def import_pygame():
    os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
    import pygame
    return pygame


def get_time_str():
    return datetime.datetime.now().strftime("%y%m%d-%H%M%S")


def setup_logger(debug=False):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format='%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s'
    )


def recursive_equal(data1, data2, need_assert=False):
    from metadrive.utils.config import Config
    if isinstance(data1, Config):
        data1 = data1.get_dict()
    if isinstance(data2, Config):
        data2 = data2.get_dict()

    if isinstance(data1, np.ndarray):
        tmp = np.asarray(data2)
        return np.all(data1 == tmp)

    if isinstance(data2, np.ndarray):
        tmp = np.asarray(data1)
        return np.all(tmp == data2)

    if isinstance(data1, dict):
        is_ins = isinstance(data2, dict)
        key_right = set(data1.keys()) == set(data2.keys())
        if need_assert:
            assert is_ins and key_right, (data1.keys(), data2.keys())
        if not (is_ins and key_right):
            return False
        ret = []
        for k in data1:
            ret.append(recursive_equal(data1[k], data2[k], need_assert=need_assert))
        return all(ret)

    elif isinstance(data1, (list, tuple)):
        len_right = len(data1) == len(data2)
        is_ins = isinstance(data2, (list, tuple))
        if need_assert:
            assert len_right and is_ins, (len(data1), len(data2), data1, data2)
        if not (is_ins and len_right):
            return False
        ret = []
        for i in range(len(data1)):
            ret.append(recursive_equal(data1[i], data2[i], need_assert=need_assert))
        return all(ret)
    elif isinstance(data1, np.ndarray):
        ret = np.isclose(data1, data2).all()
        if need_assert:
            assert ret, (type(data1), type(data2), data1, data2)
        return ret
    else:
        ret = data1 == data2
        if need_assert:
            assert ret, (type(data1), type(data2), data1, data2)
        return ret


def is_mac():
    return sys.platform == "darwin"


def is_win():
    return sys.platform == "win32"


def concat_step_infos(step_info_list):
    """We only conduct simply shallow update here!"""
    old_dict = dict()
    for new_dict in step_info_list:
        old_dict = merge_dicts(old_dict, new_dict, allow_new_keys=True, without_copy=True)
    return old_dict


# The following two functions is copied from ray/tune/utils/util.py, raise_error and pgconfig support is added by us!
def merge_dicts(old_dict, new_dict, allow_new_keys=False, without_copy=False):
    """
    Args:
        old_dict (dict, Config): Dict 1.
        new_dict (dict, Config): Dict 2.
        raise_error (bool): Whether to raise error if new key is found.

    Returns:
         dict: A new dict that is d1 and d2 deep merged.
    """
    old_dict = old_dict or dict()
    new_dict = new_dict or dict()
    if without_copy:
        merged = old_dict
    else:
        merged = copy.deepcopy(old_dict)
    _deep_update(
        merged, new_dict, new_keys_allowed=allow_new_keys, allow_new_subkey_list=[], raise_error=not allow_new_keys
    )
    return merged


def _deep_update(
    original,
    new_dict,
    new_keys_allowed=False,
    allow_new_subkey_list=None,
    override_all_if_type_changes=None,
    raise_error=True
):
    allow_new_subkey_list = allow_new_subkey_list or []
    override_all_if_type_changes = override_all_if_type_changes or []

    for k, value in new_dict.items():
        if k not in original and not new_keys_allowed:
            if raise_error:
                raise Exception("Unknown config parameter `{}` ".format(k))
            else:
                continue

        # Both orginal value and new one are dicts.
        if isinstance(original.get(k), dict) and isinstance(value, dict):
            # Check old type vs old one. If different, override entire value.
            if k in override_all_if_type_changes and \
                    "type" in value and "type" in original[k] and \
                    value["type"] != original[k]["type"]:
                original[k] = value
            # Allowed key -> ok to add new subkeys.
            elif k in allow_new_subkey_list:
                _deep_update(original[k], value, True, raise_error=raise_error)
            # Non-allowed key.
            else:
                _deep_update(original[k], value, new_keys_allowed, raise_error=raise_error)
        # Original value not a dict OR new value not a dict:
        # Override entire value.
        else:
            original[k] = value
    return original


def deprecation_warning(old, new, error=False) -> None:
    """Warns (via the `logger` object) or throws a deprecation warning/error.

    Args:
        old (str): A description of the "thing" that is to be deprecated.
        new (Optional[str]): A description of the new "thing" that replaces it.
        error (Optional[Union[bool, Exception]]): Whether or which exception to
            throw. If True, throw ValueError. If False, just warn.
            If Exception, throw that Exception.
    """
    msg = "`{}` has been deprecated.{}".format(old, (" Use `{}` instead.".format(new) if new else ""))
    if error is True:
        raise ValueError(msg)
    elif error and issubclass(error, Exception):
        raise error(msg)
    else:
        logger = logging.getLogger(__name__)
        logger.warning("DeprecationWarning: " + msg + " This will raise an error in the future!")


def get_object_from_node(node: BulletBodyNode):
    """
    Use this api to get the python object from bullet RayCast/SweepTest/CollisionCallback result
    """
    if node.getPythonTag(node.getName()) is None:
        return None
    from metadrive.engine.engine_utils import get_object
    from metadrive.engine.engine_utils import get_engine
    ret = node.getPythonTag(node.getName()).base_object_name
    is_road = node.getPythonTag(node.getName()).type_name == MetaDriveType.LANE_SURFACE_STREET
    if is_road:
        return get_engine().current_map.road_network.get_lane(ret)
    else:
        return get_object(ret)[ret]


def is_map_related_instance(obj):
    from metadrive.component.block.base_block import BaseBlock
    from metadrive.component.map.base_map import BaseMap
    return True if isinstance(obj, BaseBlock) or isinstance(obj, BaseMap) else False


def is_map_related_class(object_class):
    from metadrive.component.block.base_block import BaseBlock
    from metadrive.component.map.base_map import BaseMap
    return True if issubclass(object_class, BaseBlock) or issubclass(object_class, BaseMap) else False


def dict_recursive_remove_array(d):
    if isinstance(d, np.ndarray):
        return d.tolist()
    if isinstance(d, dict):
        for k in d.keys():
            d[k] = dict_recursive_remove_array(d[k])
    return d


def time_me(fn):
    def _wrapper(*args, **kwargs):
        start = time.time()
        ret = fn(*args, **kwargs)
        print("function: %s cost %s second" % (fn.__name__, time.time() - start))
        return ret

    return _wrapper



from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario import NuPlanScenario
from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector
from metadrive.scenario.parse_object_state import get_idm_route
from metadrive.component.lane.point_lane import PointLane
from metadrive.component.map.nuplan_map import NuPlanMap
from metadrive.constants import DEFAULT_AGENT
from metadrive.manager.base_manager import BaseManager
from metadrive.utils.nuplan.parse_object_state import parse_ego_vehicle_trajectory, parse_ego_vehicle_state


class NuPlanMapManager(BaseManager):
    """
    Build to execute the same function as NuPlanMapManager
    """
    PRIORITY = 0  # Map update has the most high priority
    DEFAULT_DATA_BUFFER_SIZE = 200

    def __init__(self):
        super(NuPlanMapManager, self).__init__()
        self.store_map = self.engine.global_config.get("store_map", False)
        self.current_map = None
        self.map_num = self.engine.data_manager.num_scenarios
        self.start = self.engine.global_config["start_scenario_index"]
        self.sdc_dest_point = None
        self.current_sdc_route = None
        self.MAP_CENTERS = self.engine.global_config["map_centers"]
        self.MAP_RADIUS = self.engine.global_config["city_map_radius"] \
            if self.engine.global_config["load_city_map"] else self.engine.global_config["scenario_radius"]

        # Now we store the whole city map which is the largest map! There are four maps in NuPlan
        self.city_maps = {}

        # It is used for storing maps separately
        self.store_map_buffer = {}

    def reset(self):
        current_map_name = self.engine.data_manager.current_scenario.map_api.map_name
        state = self.engine.data_manager.current_scenario.get_ego_state_at_iteration(0)
        scenario_center = nuplan_to_metadrive_vector([state.waypoint.x, state.waypoint.y])
        if not self.engine.global_config["load_city_map"]:
            # In this mode we don't buffer map for saving time
            if current_map_name in self.city_maps:
                self.city_maps.pop(current_map_name)
            if self.engine.data_manager.current_scenario_index in self.store_map_buffer:
                new_map = self.store_map_buffer[self.engine.data_manager.current_scenario_index]
            else:
                new_map = NuPlanMap(nuplan_center=scenario_center, map_name=current_map_name, radius=self.MAP_RADIUS)
                self.store_map_buffer[self.engine.data_manager.current_scenario_index] = new_map
            self.city_maps[current_map_name] = new_map
        else:
            # When setting loading city map, the whole map will be created
            if current_map_name in self.city_maps:
                new_map = self.city_maps[current_map_name]
            else:
                center = self.MAP_CENTERS[current_map_name]
                new_map = NuPlanMap(nuplan_center=center, map_name=current_map_name, radius=self.MAP_RADIUS)
                self.city_maps[current_map_name] = new_map
        self.load_map(new_map, scenario_center)
        self.update_ego_route()

    def filter_path(self, start_lanes, end_lanes):
        for start in start_lanes:
            for end in end_lanes:
                path = self.current_map.road_network.shortest_path(start[0].index, end[0].index)
                if len(path) > 0:
                    return (start[0].index, end[0].index)
        return None

    def spawn_object(self, object_class, *args, **kwargs):
        raise ValueError("Please create NuPlanMap instance directly without calling spawn_object function.")
        # map = self.engine.spawn_object(object_class, auto_fill_random_seed=False, *args, **kwargs)
        # self.spawned_objects[map.id] = map
        # return map

    def load_map(self, map, center):
        map.attach_to_world(center)
        self.current_map = map

    def unload_map(self, map):
        map.detach_from_world()
        self.current_map = None
        # if not self.engine.global_config["store_map"]:
        #     self.clear_objects([map.id])
        #     assert len(self.spawned_objects) == 0

    def destroy(self):
        self.current_map = None
        super(NuPlanMapManager, self).destroy()
        self.sdc_dest_point = None
        self.current_sdc_route = None

    def before_reset(self):
        # remove map from world before adding
        if self.current_map is not None:
            self.unload_map(self.current_map)
        self.sdc_dest_point = None
        self.current_sdc_route = None

    def clear_objects(self, *args, **kwargs):
        """
        As Map instance should not be recycled, we will forcefully destroy useless map instances.
        """
        return super(NuPlanMapManager, self).clear_objects(force_destroy=True, *args, **kwargs)

    def update_ego_route(self):
        """
        Ego Route is placed in map manager
        """
        scenario: NuPlanScenario = self.engine.data_manager.current_scenario

        sdc_traj = parse_ego_vehicle_trajectory(scenario.get_expert_ego_trajectory(), self.current_map.nuplan_center)
        self.current_sdc_route = get_idm_route(sdc_traj)
        init_state = parse_ego_vehicle_state(scenario.get_ego_state_at_iteration(0), self.current_map.nuplan_center)
        last_state = parse_ego_vehicle_state(
            scenario.get_ego_state_at_iteration(scenario.get_number_of_iterations() - 1), self.current_map.nuplan_center
        )

        init_position = init_state["position"]
        init_yaw = init_state["heading"]
        last_position = last_state["position"]

        self.sdc_dest_point = last_position
        self.engine.global_config.update(
            dict(
                target_vehicle_configs={
                    DEFAULT_AGENT: dict(
                        spawn_position_heading=(init_position, init_yaw),
                        spawn_velocity=init_state["velocity"],
                        spawn_velocity_car_frame=True,
                    )
                }
            )
        )



from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv


def test_safe_env(vis=False):
    config = {"num_scenarios": 100, "start_seed": 75}
    if vis:
        config["vehicle_config"] = {"show_line_to_navi_mark": True}
        config["use_render"] = True
        config["manual_control"] = True
        config["controller"] = "joystick"

    env = SafeMetaDriveEnv(config)
    try:
        o, _ = env.reset()
        total_cost = 0
        for ep in range(5):
            for i in range(1, 100):
                o, r, tm, tc, info = env.step([0, 1])
                total_cost += info["cost"]
                assert env.observation_space.contains(o)
                if tm or tc:
                    total_cost = 0
                    # print("Reset")
                    env.reset()
        env.close()
    finally:
        env.close()


if __name__ == '__main__':
    test_safe_env(vis=True)



import logging

from metadrive.version import VERSION, asset_version
import os
from metadrive.pull_asset import pull_asset
from metadrive.constants import RENDER_MODE_NONE, RENDER_MODE_OFFSCREEN, RENDER_MODE_ONSCREEN
import pickle
import time
from collections import OrderedDict
from typing import Callable, Optional, Union, List, Dict, AnyStr

import numpy as np
from panda3d.core import NodePath, Vec3

from metadrive.base_class.randomizable import Randomizable
from metadrive.engine.core.engine_core import EngineCore
from metadrive.engine.interface import Interface
from metadrive.manager.base_manager import BaseManager
from metadrive.utils import concat_step_infos
from metadrive.utils.utils import is_map_related_class

logger = logging.getLogger(__name__)


class BaseEngine(EngineCore, Randomizable):
    """
    Due to the feature of Panda3D, BaseEngine should only be created once(Singleton Pattern)
    It is a pure game engine, which is not task-specific, while BaseEngine connects the
    driving task and the game engine modified from Panda3D Engine.
    """
    singleton = None
    global_random_seed = None

    def __init__(self, global_config):
        self.try_pull_asset()
        EngineCore.__init__(self, global_config)
        Randomizable.__init__(self, self.global_random_seed)
        self.episode_step = 0
        BaseEngine.singleton = self
        self.interface = Interface(self)

        # managers
        self.task_manager = self.taskMgr  # use the inner TaskMgr of Panda3D as MetaDrive task manager
        self._managers = OrderedDict()

        # for recovering, they can not exist together
        self.record_episode = False
        self.replay_episode = False
        self.only_reset_when_replay = False
        # self.accept("s", self._stop_replay)

        # cull scene
        self.cull_scene = self.global_config["cull_scene"]

        # add camera or not
        self.main_camera = self.setup_main_camera()

        self._spawned_objects = dict()
        self._object_policies = dict()
        self._object_tasks = dict()

        # the clear function is a fake clear, objects cleared is stored for future use
        self._dying_objects = dict()

        # store external actions
        self.external_actions = None

        # topdown renderer
        self._top_down_renderer = None

        # lanes debug
        self.lane_coordinates_debug_node = None

        # warm up
        self.warmup()

        # curriculum reset
        self._max_level = self.global_config.get("curriculum_level", 1)
        self._current_level = 0
        self._num_scenarios_per_level = int(self.global_config.get("num_scenarios", 0) / self._max_level)

    def add_policy(self, object_id, policy_class, *args, **kwargs):
        policy = policy_class(*args, **kwargs)
        self._object_policies[object_id] = policy
        if self.record_episode:
            assert self.record_manager is not None, "No record manager"
            self.record_manager.add_policy_info(object_id, policy_class, *args, **kwargs)
        return policy

    def add_task(self, object_id, task):
        self._object_tasks[object_id] = task

    def get_policy(self, object_id):
        """
        Return policy of specific object with id
        :param object_id: a filter function, only return objects satisfying this condition
        :return: policy
        """
        if object_id in self._object_policies:
            return self._object_policies[object_id]
        else:
            # print("Can not find the policy for object(id: {})".format(object_id))
            return None

    def get_task(self, object_id):
        """
        Return task of specific object with id
        :param object_id: a filter function, only return objects satisfying this condition
        :return: task
        """
        assert object_id in self._object_tasks, "Can not find the task for object(id: {})".format(object_id)
        return self._object_tasks[object_id]

    def has_policy(self, object_id, policy_cls=None):
        if policy_cls is None:
            return True if object_id in self._object_policies else False
        else:
            return True if object_id in self._object_policies and isinstance(
                self._object_policies[object_id], policy_cls
            ) else False

    def has_task(self, object_id):
        return True if object_id in self._object_tasks else False

    def spawn_object(
        self, object_class, pbr_model=True, force_spawn=False, auto_fill_random_seed=True, record=True, **kwargs
    ):
        """
        Call this func to spawn one object
        :param object_class: object class
        :param pbr_model: if the visualization model is pbr model
        :param force_spawn: spawn a new object instead of fetching from _dying_objects list
        :param auto_fill_random_seed: whether to set random seed using purely random integer
        :param record: record the spawn information
        :param kwargs: class init parameters
        :return: object spawned
        """
        if ("random_seed" not in kwargs) and auto_fill_random_seed:
            kwargs["random_seed"] = self.generate_seed()
        if force_spawn or object_class.__name__ not in self._dying_objects or len(
                self._dying_objects[object_class.__name__]) == 0:
            obj = object_class(**kwargs)
        else:
            obj = self._dying_objects[object_class.__name__].pop()
            obj.reset(**kwargs)
            if not is_map_related_class(object_class) and ("name" not in kwargs or kwargs["name"] is None):
                obj.random_rename()

        if "name" in kwargs and kwargs["name"] is not None:
            assert kwargs["name"] == obj.name == obj.id
        if "id" in kwargs and kwargs["name"] is not None:
            assert kwargs["id"] == obj.id == obj.name

        if self.global_config["record_episode"] and not self.replay_episode and record:
            self.record_manager.add_spawn_info(obj, object_class, kwargs)
        self._spawned_objects[obj.id] = obj
        obj.attach_to_world(self.pbr_worldNP if pbr_model else self.worldNP, self.physics_world)
        return obj

    def get_objects(self, filter: Optional[Union[Callable, List]] = None):
        """
        Return objects spawned, default all objects. Filter_func will be applied on all objects.
        It can be a id list or a function
        Since we don't expect a iterator, and the number of objects is not so large, we don't use built-in filter()
        :param filter: a filter function, only return objects satisfying this condition
        :return: return all objects or objects satisfying the filter_func
        """
        if filter is None:
            return self._spawned_objects
        elif isinstance(filter, (list, tuple)):
            return {id: self._spawned_objects[id] for id in filter}
        elif callable(filter):
            res = dict()
            for id, obj in self._spawned_objects.items():
                if filter(obj):
                    res[id] = obj
            return res
        else:
            raise ValueError("filter should be a list or a function")

    def get_policies(self):
        """
        Return a mapping from object ID to policy instance.
        """
        return self._object_policies

    def get_object(self, object_id):
        return self.get_objects([object_id])

    def clear_objects(self, filter: Optional[Union[Callable, List]], force_destroy=False, record=True):
        """
        Destroy all self-generated objects or objects satisfying the filter condition
        Since we don't expect a iterator, and the number of objects is not so large, we don't use built-in filter()
        If force_destroy=True, we will destroy this element instead of storing them for next time using

        filter: A list of object ids or a function returning a list of object id
        """
        force_destroy_this_obj = True if force_destroy or self.global_config["force_destroy"] else False

        if isinstance(filter, (list, tuple)):
            exclude_objects = {obj_id: self._spawned_objects[obj_id] for obj_id in filter}
        elif callable(filter):
            exclude_objects = dict()
            for id, obj in self._spawned_objects.items():
                if filter(obj):
                    exclude_objects[id] = obj
        else:
            raise ValueError("filter should be a list or a function")
        for id, obj in exclude_objects.items():
            self._spawned_objects.pop(id)
            if id in self._object_tasks:
                self._object_tasks.pop(id)
            if id in self._object_policies:
                policy = self._object_policies.pop(id)
                policy.destroy()
            if force_destroy_this_obj:
                obj.destroy()
            else:
                obj.detach_from_world(self.physics_world)

                # We might want to remove some episode-relevant information when recycling some objects
                if hasattr(obj, "before_reset"):
                    obj.before_reset()

                if obj.class_name not in self._dying_objects:
                    self._dying_objects[obj.class_name] = []
                # We have a limit for buffering objects
                if len(self._dying_objects[obj.class_name]) < self.global_config["num_buffering_objects"]:
                    self._dying_objects[obj.class_name].append(obj)
                else:
                    obj.destroy()
            if self.global_config["record_episode"] and not self.replay_episode and record:
                self.record_manager.add_clear_info(obj)
        return exclude_objects.keys()

    def clear_object_if_possible(self, obj, force_destroy):
        if isinstance(obj, dict):
            return
        if obj in self._spawned_objects:
            self.clear_objects([obj], force_destroy=force_destroy)
        if force_destroy and \
                obj.class_name in self._dying_objects and \
                obj in self._dying_objects[obj.class_name]:
            self._dying_objects[obj.class_name].remove(obj)
            if hasattr(obj, "destroy"):
                obj.destroy()
        del obj

    def reset(self):
        """
        Clear and generate the whole scene
        """
        # initialize
        self._episode_start_time = time.time()
        self.episode_step = 0
        if self.global_config["debug_physics_world"]:
            self.addTask(self.report_body_nums, "report_num")

        # Update record replay
        self.replay_episode = True if self.global_config["replay_episode"] is not None else False
        self.record_episode = self.global_config["record_episode"]
        self.only_reset_when_replay = self.global_config["only_reset_when_replay"]

        _debug_memory_usage = False

        if _debug_memory_usage:

            def process_memory():
                import psutil
                import os
                process = psutil.Process(os.getpid())
                mem_info = process.memory_info()
                return mem_info.rss

            cm = process_memory()

        # reset manager
        for manager_name, manager in self._managers.items():
            # clean all manager
            manager.before_reset()

            if _debug_memory_usage:
                lm = process_memory()
                if lm - cm != 0:
                    print("{}: Before Reset! Mem Change {:.3f}MB".format(manager_name, (lm - cm) / 1e6))
                cm = lm

        self._object_clean_check()

        for manager_name, manager in self.managers.items():
            if self.replay_episode and self.only_reset_when_replay and manager is not self.replay_manager:
                # The scene will be generated from replay manager in only reset replay mode
                continue
            manager.reset()

            if _debug_memory_usage:
                lm = process_memory()
                if lm - cm != 0:
                    print("{}: Reset! Mem Change {:.3f}MB".format(manager_name, (lm - cm) / 1e6))
                cm = lm

        for manager_name, manager in self.managers.items():
            manager.after_reset()

            if _debug_memory_usage:
                lm = process_memory()
                if lm - cm != 0:
                    print("{}: After Reset! Mem Change {:.3f}MB".format(manager_name, (lm - cm) / 1e6))
                cm = lm

        # reset cam
        if self.main_camera is not None:
            self.main_camera.reset()
            if hasattr(self, "agent_manager"):
                bev_cam = self.main_camera.is_bird_view_camera() and self.main_camera.current_track_vehicle is not None
                vehicles = list(self.agents.values())
                current_track_vehicle = vehicles[0]
                self.main_camera.set_follow_lane(self.global_config["use_chase_camera_follow_lane"])
                self.main_camera.track(current_track_vehicle)
                if bev_cam:
                    self.main_camera.stop_track()
                    self.main_camera.set_bird_view_pos(current_track_vehicle.position)

                # if self.global_config["is_multi_agent"]:
                #     self.main_camera.stop_track(bird_view_on_current_position=False)

        # reset terrain
        center_p = self.current_map.get_center_point() if self.current_map else [0, 0]
        self.terrain.reset(center_p)
        if self.sky_box is not None:
            self.sky_box.set_position(center_p)

        self.taskMgr.step()

        # refresh graphics to support multi-thread rendering, avoiding bugs like shadow disappearance at first frame
        for _ in range(5):
            self.graphicsEngine.renderFrame()

    def before_step(self, external_actions: Dict[AnyStr, np.array]):
        """
        Entities make decision here, and prepare for step
        All entities can access this global manager to query or interact with others
        :param external_actions: Dict[agent_id:action]
        :return:
        """
        self.episode_step += 1
        step_infos = {}
        self.external_actions = external_actions
        for manager in self.managers.values():
            new_step_infos = manager.before_step()
            step_infos = concat_step_infos([step_infos, new_step_infos])
        return step_infos

    def step(self, step_num: int = 1) -> None:
        """
        Step the dynamics of each entity on the road.
        :param step_num: Decision of all entities will repeat *step_num* times
        """
        for i in range(step_num):
            # simulate or replay
            for name, manager in self.managers.items():
                if name != "record_manager":
                    manager.step()
            self.step_physics_world()
            # the recording should happen after step physics world
            if "record_manager" in self.managers and i < step_num - 1:
                # last recording should be finished in after_step(), as some objects may be created in after_step.
                # We repeat run simulator ```step_num``` frames, and record after each frame.
                # The recording of last frame is actually finished when all managers finish the ```after_step()```
                # function. So the recording for the last time should be done after that.
                # An example is that in ```PGTrafficManager``` we may create new vehicles in
                # ```after_step()``` of the traffic manager. Therefore, we can't record the frame before that.
                # These new cars' states can be recorded only if we run ```record_managers.step()```
                # after the creation of new cars and then can be recorded in ```record_managers.after_step()```
                self.record_manager.step()

            if self.force_fps.real_time_simulation and i < step_num - 1:
                self.task_manager.step()
        #  panda3d render and garbage collecting loop
        self.task_manager.step()
        if self.on_screen_message is not None:
            self.on_screen_message.render()

    def after_step(self, *args, **kwargs) -> Dict:
        """
        Update states after finishing movement
        :return: if this episode is done
        """

        step_infos = {}
        if self.record_episode:
            assert list(self.managers.keys())[-1] == "record_manager", "Record Manager should have lowest priority"
        for manager in self.managers.values():
            new_step_info = manager.after_step(*args, **kwargs)
            step_infos = concat_step_infos([step_infos, new_step_info])
        self.interface.after_step()

        # === Option 1: Set episode_step to "num of calls to env.step"
        # We want to make sure that the episode_step is always aligned to the "number of calls to env.step"
        # So if this function is called in env.reset, we will not increment episode_step.
        # if call_from_reset:
        #     pass
        # else:
        #     self.episode_step += 1

        # === Option 2: Following old code.
        # Note that this function will be called in _get_reset_return.
        # Therefore, after reset the episode_step is immediately goes to 1
        # even if no env.step is called.

        # Episode_step should be increased before env.step(). I moved it to engine.before_step() now.

        # cull distant blocks
        # poses = [v.position for v in self.agent_manager.active_agents.values()]
        # if self.cull_scene and False:
        #     SceneCull.cull_distant_blocks(self, self.current_map.blocks, poses, self.global_config["max_distance"])
        return step_infos

    def dump_episode(self, pkl_file_name=None) -> None:
        """Dump the data of an episode."""
        assert self.record_manager is not None
        episode_state = self.record_manager.get_episode_metadata()
        if pkl_file_name is not None:
            with open(pkl_file_name, "wb+") as file:
                pickle.dump(episode_state, file)
        return episode_state

    def close(self):
        """
        Note:
        Instead of calling this func directly, close Engine by using engine_utils.close_engine
        """
        if len(self._managers) > 0:
            for name, manager in self._managers.items():
                setattr(self, name, None)
                if manager is not None:
                    manager.destroy()
        # clear all objects in spawned_object
        # self.clear_objects([id for id in self._spawned_objects.keys()])
        for id, obj in self._spawned_objects.items():
            if id in self._object_policies:
                self._object_policies.pop(id).destroy()
            if id in self._object_tasks:
                self._object_tasks.pop(id).destroy()
            obj.destroy()
        for cls, pending_obj in self._dying_objects.items():
            for obj in pending_obj:
                obj.destroy()
        if self.main_camera is not None:
            self.main_camera.destroy()
        self.interface.destroy()
        self.close_world()

        if self._top_down_renderer is not None:
            self._top_down_renderer.close()
            del self._top_down_renderer
            self._top_down_renderer = None

    def __del__(self):
        logging.debug("{} is destroyed".format(self.__class__.__name__))

    def _stop_replay(self):
        raise DeprecationWarning
        if not self.IN_REPLAY:
            return
        self.STOP_REPLAY = not self.STOP_REPLAY

    def register_manager(self, manager_name: str, manager: BaseManager):
        """
        Add a manager to BaseEngine, then all objects can communicate with this class
        :param manager_name: name shouldn't exist in self._managers and not be same as any class attribute
        :param manager: subclass of BaseManager
        """
        assert manager_name not in self._managers, "Manager already exists in BaseEngine, Use update_manager() to " \
                                                   "overwrite"
        assert not hasattr(self, manager_name), "Manager name can not be same as the attribute in BaseEngine"
        self._managers[manager_name] = manager
        setattr(self, manager_name, manager)
        self._managers = OrderedDict(sorted(self._managers.items(), key=lambda k_v: k_v[-1].PRIORITY))

    def seed(self, random_seed):
        start_seed = self.gets_start_index()
        random_seed = ((random_seed - start_seed) % self._num_scenarios_per_level) + start_seed
        random_seed += self._current_level * self._num_scenarios_per_level
        self.global_random_seed = random_seed
        super(BaseEngine, self).seed(random_seed)
        for mgr in self._managers.values():
            mgr.seed(random_seed)

    def gets_start_index(self):
        start_seed = self.global_config.get("start_seed", None)
        start_scenario_index = self.global_config.get("start_scenario_index", None)
        if start_seed is None:
            assert start_scenario_index is not None
            return start_scenario_index
        else:
            assert start_seed is not None
            return start_seed

    @property
    def max_level(self):
        return self._max_level

    @property
    def current_level(self):
        return self._current_level

    def level_up(self):
        old_level = self._current_level
        self._current_level = min(self._current_level + 1, self._max_level - 1)
        if old_level != self._current_level:
            self.seed(self.current_seed + self._num_scenarios_per_level)

    @property
    def num_scenarios_per_level(self):
        return self._num_scenarios_per_level

    @property
    def current_map(self):
        if self.replay_episode:
            return self.replay_manager.current_map
        else:
            if hasattr(self, "map_manager"):
                return self.map_manager.current_map
            else:
                raise ValueError("No mapmanager in {}".format(self.managers))

    @property
    def current_track_vehicle(self):
        if self.main_camera is not None:
            return self.main_camera.current_track_vehicle
        elif "default_agent" in self.agents:
            return self.agents["default_agent"]
        else:
            return None

    @property
    def agents(self):
        if not self.replay_episode:
            return self.agent_manager.active_agents
        else:
            return self.replay_manager.replay_agents

    def setup_main_camera(self):
        from metadrive.engine.core.main_camera import MainCamera
        # Not we should always enable main camera if image obs is required! Or RGBCamera will return incorrect result
        if self.mode == RENDER_MODE_ONSCREEN or (RENDER_MODE_OFFSCREEN and
                                                 self.global_config["vehicle_config"]["image_source"] == "main_camera"):
            return MainCamera(self, self.global_config["camera_height"], self.global_config["camera_dist"])
        else:
            return None

    @property
    def current_seed(self):
        return self.global_random_seed

    @property
    def global_seed(self):
        return self.global_random_seed

    def _object_clean_check(self):
        if self.global_config["debug"]:
            from metadrive.component.vehicle.base_vehicle import BaseVehicle
            from metadrive.component.static_object.traffic_object import TrafficObject
            for manager in self._managers.values():
                assert len(manager.spawned_objects) == 0

            objs_need_to_release = self.get_objects(
                filter=lambda obj: isinstance(obj, BaseVehicle) or isinstance(obj, TrafficObject)
            )
            assert len(
                objs_need_to_release) == 0, "You should clear all generated objects by using engine.clear_objects " \
                                            "in each manager.before_step()"

    def update_manager(self, manager_name: str, manager: BaseManager, destroy_previous_manager=True):
        """
        Update an existing manager with a new one
        :param manager_name: existing manager name
        :param manager: new manager
        """
        assert manager_name in self._managers, "You may want to call register manager, since {} is not in engine".format(
            manager_name
        )
        existing_manager = self._managers.pop(manager_name)
        if destroy_previous_manager:
            existing_manager.destroy()
        self._managers[manager_name] = manager
        setattr(self, manager_name, manager)
        self._managers = OrderedDict(sorted(self._managers.items(), key=lambda k_v: k_v[-1].PRIORITY))

    @property
    def managers(self):
        # whether to froze other managers
        return {"replay_manager": self.replay_manager} if self.replay_episode and not \
            self.only_reset_when_replay else self._managers

    def change_object_name(self, obj, new_name):
        raise DeprecationWarning("This function is too dangerous to be used")
        """
        Change the name of one object, Note: it may bring some bugs if abusing
        """
        obj = self._spawned_objects.pop(obj.name)
        self._spawned_objects[new_name] = obj

    def object_to_agent(self, obj_name):
        if self.replay_episode:
            return self.replay_manager.current_frame.object_to_agent(obj_name)
        else:
            return self.agent_manager.object_to_agent(obj_name)

    def agent_to_object(self, agent_name):
        if self.replay_episode:
            return self.replay_manager.current_frame.agent_to_object(agent_name)
        else:
            return self.agent_manager.agent_to_object(agent_name)

    def render_topdown(self, text, *args, **kwargs):
        if self._top_down_renderer is None:
            from metadrive.obs.top_down_renderer import TopDownRenderer
            self._top_down_renderer = TopDownRenderer(*args, **kwargs)
        return self._top_down_renderer.render(text, *args, **kwargs)

    def get_window_image(self, return_bytes=False):
        window_count = self.graphicsEngine.getNumWindows() - 1
        texture = self.graphicsEngine.getWindow(window_count).getDisplayRegion(0).getScreenshot()

        assert texture.getXSize() == self.global_config["window_size"][0], (
            texture.getXSize(), texture.getYSize(), self.global_config["window_size"]
        )
        assert texture.getYSize() == self.global_config["window_size"][1], (
            texture.getXSize(), texture.getYSize(), self.global_config["window_size"]
        )

        image_bytes = texture.getRamImage().getData()

        if return_bytes:
            return image_bytes, (texture.getXSize(), texture.getYSize())

        img = np.frombuffer(image_bytes, dtype=np.uint8)
        img = img.reshape((texture.getYSize(), texture.getXSize(), 4))
        img = img[::-1]  # Flip vertically
        img = img[..., :-1]  # Discard useless alpha channel
        img = img[..., ::-1]  # Correct the colors

        return img

    def show_lane_coordinates(self, lanes):
        if self.lane_coordinates_debug_node is not None:
            self.lane_coordinates_debug_node.detachNode()
            self.lane_coordinates_debug_node.removeNode()

        self.lane_coordinates_debug_node = NodePath("Lane Coordinates debug")
        for lane in lanes:
            long_start = lateral_start = lane.position(0, 0)
            lateral_end = lane.position(0, 2)

            long_end = long_start + lane.heading_at(0) * 4
            np_y = self.draw_line_3d(Vec3(*long_start, 0), Vec3(*long_end, 0), color=[0, 1, 0, 1], thickness=2)
            np_x = self.draw_line_3d(Vec3(*lateral_start, 0), Vec3(*lateral_end, 0), color=[1, 0, 0, 1], thickness=2)
            np_x.reparentTo(self.lane_coordinates_debug_node)
            np_y.reparentTo(self.lane_coordinates_debug_node)
        self.lane_coordinates_debug_node.reparentTo(self.worldNP)

    def remove_show_lane_coordinates(self):
        if self.lane_coordinates_debug_node is not None:
            self.lane_coordinates_debug_node.detachNode()
            self.lane_coordinates_debug_node.removeNode()

    def warmup(self):
        """
        This function automatically initialize models/objects. It can prevent the lagging when creating some objects
        for the first time.
        """
        if self.global_config["preload_models"]:
            from metadrive.component.traffic_participants.pedestrian import Pedestrian
            from metadrive.component.traffic_light.base_traffic_light import BaseTrafficLight
            from metadrive.component.static_object.traffic_object import TrafficBarrier
            from metadrive.component.static_object.traffic_object import TrafficCone
            Pedestrian.init_pedestrian_model()
            warm_up_pedestrian = self.spawn_object(Pedestrian, position=[0, 0], heading_theta=0, record=False)
            warm_up_light = self.spawn_object(BaseTrafficLight, lane=None, position=[0, 0], record=False)
            barrier = self.spawn_object(TrafficBarrier, position=[0, 0], heading_theta=0, record=False)
            cone = self.spawn_object(TrafficCone, position=[0, 0], heading_theta=0, record=False)
            for vel in Pedestrian.SPEED_LIST:
                warm_up_pedestrian.set_velocity([1, 0], vel - 0.1)
                self.taskMgr.step()
            self.clear_objects([warm_up_pedestrian.id, warm_up_light.id, barrier.id, cone.id], record=False)
            warm_up_pedestrian = None
            warm_up_light = None
            barrier = None
            cone = None

    @staticmethod
    def try_pull_asset():
        from metadrive.engine.asset_loader import AssetLoader
        msg = "Assets folder doesn't exist. Begin to download assets..."
        if not os.path.exists(AssetLoader.asset_path):
            AssetLoader.logger.warning(msg)
            pull_asset(False)
        else:
            if asset_version() != VERSION:
                AssetLoader.logger.warning(
                    "Assets version mismatch! Current: {}, Expected: {}. "
                    "Update the assets by `python -m metadrive.pull_asset --update'".format(asset_version(), VERSION)
                )
            else:
                AssetLoader.logger.info("Assets version: {}".format(VERSION))


if __name__ == "__main__":
    from metadrive.envs.base_env import BASE_DEFAULT_CONFIG

    BASE_DEFAULT_CONFIG["use_render"] = True
    BASE_DEFAULT_CONFIG["show_interface"] = False
    BASE_DEFAULT_CONFIG["render_pipeline"] = True
    world = BaseEngine(BASE_DEFAULT_CONFIG)

    from metadrive.engine.asset_loader import AssetLoader

    car_model = world.loader.loadModel(AssetLoader.file_path("models", "vehicle", "lada", "vehicle.gltf"))
    car_model.reparentTo(world.render)
    car_model.set_pos(0, 0, 190)
    # world.render_pipeline.prepare_scene(env.engine.render)

    world.run()



import time

from metadrive.component.map.scenario_map import ScenarioMap
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import initialize_engine, close_engine
from metadrive.envs.real_data_envs.waymo_env import WaymoEnv
from metadrive.manager.waymo_data_manager import WaymoDataManager
from metadrive.tests.test_functionality.test_memory_leak_engine import process_memory


def test_waymo_env_memory_leak(num_reset=100):
    env = WaymoEnv(dict(num_scenarios=2, sequential_seed=True, store_map=True))

    try:
        ct = time.time()
        cm = process_memory()
        for t in range(num_reset):
            lt = time.time()
            env.reset()
            nlt = time.time()
            lm = process_memory()
            # print(
            #     "After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,}".format(
            #         t + 1, nlt - lt, nlt - ct, lm - cm
            #     )
            # )
        assert lm - cm < 1024 * 1024 * 120, "We expect will cause ~120MB memory leak."

    finally:
        env.close()


def test_waymo_map_memory_leak():
    default_config = WaymoEnv.default_config()
    default_config["data_directory"] = AssetLoader.file_path("waymo", return_raw_style=False)
    default_config["num_scenarios"] = 1

    try:
        close_engine()
        engine = initialize_engine(default_config)

        ct = time.time()
        cm = process_memory()
        last_mem = 0.0

        lt = time.time()

        engine.data_manager = WaymoDataManager()
        engine.seed(0)

        lm = process_memory()
        nlt = time.time()
        # print("After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,}".format(0, nlt - lt, nlt - ct, lm - cm))

        for t in range(10):
            lt = time.time()

            map = ScenarioMap(map_index=0)
            del map

            # map.play()

            nlt = time.time()
            lm = process_memory()
            # print(
            #     "After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,}".format(
            #         t + 1, nlt - lt, nlt - ct, lm - cm
            #     )
            # )
            # if t > 5:
            #     assert abs((lm - cm) - last_mem) < 1024  # Memory should not have change > 1KB
            last_mem = lm - cm

        assert lm - cm < 1024 * 1024 * 40, "We expect will cause ~33MB memory leak."

    finally:
        close_engine()


if __name__ == "__main__":

    # https://code.activestate.com/recipes/65333/

    # import gc

    # def dump_garbage():
    #     """
    #     show us what's the garbage about
    #     """
    #
    #     # force collection
    #     # print("\nGARBAGE:")
    #     gc.collect()
    #
    #     # print("\nGARBAGE OBJECTS:")
    #     res = []
    #     for x in gc.garbage:
    #         s = str(x)
    #         if len(s) > 80:
    #             s = s[:80]
    #         # print(type(x), "\n  ", s)
    #         res.append([type(x), s, x])
    #     return res

    # gc.enable()
    # gc.set_debug(gc.DEBUG_LEAK)

    test_waymo_env_memory_leak(num_reset=300)

    test_waymo_map_memory_leak()

    # show the dirt ;-)
    # ret = dump_garbage()

    # # print(ret)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division, print_function
from math import exp, log

from panda3d.core import PTAFloat, PTALVecBase3f

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.util.smooth_connected_curve import SmoothConnectedCurve
from metadrive.render_pipeline.rpcore.pluginbase.setting_types import make_setting_from_factory

__all__ = ["make_daysetting_from_data"]


def make_daysetting_from_data(data):
    """ Constructs a new setting from a given dataset. This method will automatically
    instantiate a new class matching the type of the given dataset. It will fill
    all values provided by the dataset and then return the created instance """
    factory = {"color": ColorType, "scalar": ScalarType}
    return make_setting_from_factory(data, factory)


class BaseType(RPObject):
    """ Base setting type for all setting types """
    def __init__(self, data):
        self.type = data.pop("type")
        self.label = data.pop("label").strip()
        self.description = data.pop("description").strip()
        self.curves = []

        RPObject.__init__(self, "dsetting:{}".format(self.label))

    def get_value_at(self, offset):
        """ Returns the unscaled value at the given day time offset """
        if len(self.curves) == 1:
            return self.curves[0].get_value(offset)
        return tuple(curve.get_value(offset) for curve in self.curves)

    def get_scaled_value_at(self, offset):
        """ Returns the scaled value at a given day time offset """
        return self.get_scaled_value(self.get_value_at(offset))

    def get_scaled_value(self, value):
        """ Returns the scaled value from a given normalized value """
        raise NotImplementedError()

    def set_control_points(self, control_points):
        """ Sets the control points on the curves. """
        for curve_index, points in enumerate(control_points):
            self.curves[curve_index].control_points = points

    def serialize(self):
        """ Serializes the setting to a yaml string """
        values = ','.join(i.serialize() for i in self.curves)
        return "[{}]".format(values)


class ScalarType(BaseType):
    """ Setting type storing a single scalar """

    glsl_type = "float"
    pta_type = PTAFloat

    def __init__(self, data):
        BaseType.__init__(self, data)
        self.unit = data.pop("unit", "none")
        self.minvalue, self.maxvalue = data.pop("range")
        self.logarithmic_factor = data.pop("logarithmic_factor", 1.0)
        if self.unit not in ("degree", "meter", "percent", "klux", "none"):
            raise Exception("Invalid unit type: {}".format(self.unit))

        self.default = self.get_linear_value(data.pop("default"))

        self.curves.append(SmoothConnectedCurve())
        self.curves[0].set_single_value(self.default)

    def format(self, value):
        """ Formats a given value, attaching the appropriate metric unit """
        metric = {"degree": u'\N{DEGREE SIGN}', "percent": u'%', "meter": u'm', "klux": u' L', "none": ''}[self.unit]
        if self.unit == "percent":
            value *= 100.0
        return u"{:3.1f}{}".format(value, metric)

    def get_scaled_value(self, value):
        """ Scales a linear value """
        if self.logarithmic_factor != 1.0:
            exp_mult = exp(self.logarithmic_factor * value * 4.0) - 1
            exp_div = exp(self.logarithmic_factor * 4.0) - 1
            return exp_mult / exp_div * (self.maxvalue - self.minvalue) + self.minvalue
        else:
            return value * (self.maxvalue - self.minvalue) + self.minvalue

    def get_linear_value(self, scaled_value):
        """ Linearizes a scaled value """
        result = (scaled_value - self.minvalue) / (self.maxvalue - self.minvalue)
        if self.logarithmic_factor != 1.0:
            result *= exp(self.logarithmic_factor * 4.0) - 1
            result = log(result + 1.0) / (4.0 * self.logarithmic_factor)
        return result


class ColorType(BaseType):
    """ Setting type storing a RGB color triple """

    glsl_type = "vec3"
    pta_type = PTALVecBase3f

    def __init__(self, data):
        BaseType.__init__(self, data)
        self.default = self.get_linear_value(data.pop("default"))
        colors = ((255, 0, 0), (0, 255, 0), (0, 0, 255))
        for i in range(3):
            curve = SmoothConnectedCurve()
            curve.set_single_value(self.default[i])
            curve.color = colors[i]
            self.curves.append(curve)

    def format(self, value):
        return "{:3d}, {:3d}, {:3d}".format(*(int(i) for i in value))

    def get_scaled_value(self, value):
        return tuple(i * 255.0 for i in value)

    def get_linear_value(self, scaled_value):
        return tuple(i / 255.0 for i in scaled_value)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.material import get_from_node


class MaterialMixin(object):
    def get_metallic(self, material, shader):
        if shader.type in ('BSDF_GLASS', 'BSDF_ANISOTROPIC'):
            return 1

        # Math [Value] -> [Metallic] Principled BSDF
        # math_node = get_from_node(
        #     material.node_tree, 'MATH', to_node=shader,
        #     from_socket_name='Value', to_socket_name='Metallic')
        # if math_node:
        #     for input_ in math_node.inputs:
        #         if input_.name == 'Value' and not input_.is_linked:
        #             if input_.default_value < 0.5:
        #                 return 0
        #             else:
        #                 return 1

        # elif not shader.inputs['Metallic'].is_linked:
        #     return shader.inputs['Metallic'].default_value

        if shader.inputs['Metallic'].is_linked:
            return 0  # metallic map is not supported in RP -> disable
        else:
            return shader.inputs['Metallic'].default_value

    def get_roughness(self, material, shader):
        # Math [Value] -> [Roughness] Principled BSDF
        math_node = get_from_node(
            material.node_tree, 'MATH', to_node=shader, from_socket_name='Value', to_socket_name='Roughness'
        )
        if math_node:
            for input_ in math_node.inputs:
                if input_.name == 'Value' and not input_.is_linked:
                    return input_.default_value
        else:
            return shader.inputs['Roughness'].default_value

    def get_emission(self, material, shader):
        # if not shader.inputs['Emission'].is_linked and not shader.inputs['Emission Strength'].is_linked:
        #     r, g, b, *_ = shader.inputs['Emission'].default_value
        #     e = shader.inputs['Emission Strength'].default_value
        #     return (r * e, g * e, b * e)

        # Mix RGB [Color] -> [Emission] Principled BSDF
        mix_node = get_from_node(
            material.node_tree, 'MIX_RGB', to_node=shader, from_socket_name='Color', to_socket_name='Emission'
        )
        if mix_node:
            for input_ in mix_node.inputs:
                if input_.name.startswith('Color') and not input_.is_linked:
                    return input_.default_value
        else:
            return shader.inputs['Emission'].default_value

    def get_normal_strength(self, material, shader):
        # Normal Map [Normal] -> [Normal] Principled BSDF
        normal_map = get_from_node(
            material.node_tree, 'NORMAL_MAP', to_node=shader, from_socket_name='Normal', to_socket_name='Normal'
        )
        if normal_map:
            return normal_map.inputs['Strength'].default_value
        else:
            return 1



import time

import numpy as np

from metadrive import MetaDriveEnv
from metadrive.constants import DEFAULT_AGENT
from metadrive.examples import expert, get_terminal_state


def _evaluate(env_config, num_episode, has_traffic=True, need_on_same_lane=True):
    s = time.time()
    np.random.seed(0)
    env_config["random_spawn_lane_index"] = False
    env = MetaDriveEnv(env_config)
    lane_idx_need_to_stay = 0
    try:
        obs, _ = env.reset()
        lidar_success = False
        success_list, reward_list, ep_reward, ep_len, ep_count = [], [], 0, 0, 0
        while ep_count < num_episode:
            action = expert(env.vehicle, deterministic=True)
            obs, reward, terminated, truncated, info = env.step(action)
            if need_on_same_lane:
                assert lane_idx_need_to_stay == env.vehicle.lane_index[-1], "Not one the same lane"
            # double check lidar
            if env.config["use_render"]:
                env.render(text={"lane_index": env.vehicle.lane_index, "step": env.episode_step})
            lidar = [True if p == 1.0 else False for p in env.observations[DEFAULT_AGENT].cloud_points]
            if not all(lidar):
                lidar_success = True
            ep_reward += reward
            ep_len += 1
            if terminated or truncated:
                ep_count += 1
                success_list.append(1 if get_terminal_state(info) == "Success" else 0)
                reward_list.append(ep_reward)
                ep_reward = 0
                ep_len = 0
                env.config["target_vehicle_configs"]["default_agent"]["spawn_lane_index"] = (
                    ">", ">>", len(reward_list) % 3
                )
                lane_idx_need_to_stay = len(reward_list) % 3
                obs, _ = env.reset()
                if has_traffic:
                    assert lidar_success
                lidar_success = False
        env.close()
        t = time.time() - s
        ep_reward_mean = sum(reward_list) / len(reward_list)
        success_rate = sum(success_list) / len(success_list)
        # print(
        #     f"Finish {ep_count} episodes in {t:.3f} s. Episode reward: {ep_reward_mean}, success rate: {success_rate}."
        # )
    finally:
        env.close()
    return ep_reward_mean, success_rate


def test_expert_with_traffic(use_render=False):
    ep_reward, success_rate = _evaluate(
        dict(
            num_scenarios=1,
            map="CCC",
            start_seed=2,
            random_traffic=False,
            # debug_static_world=True,
            # debug=True,
            use_render=use_render,
            vehicle_config=dict(show_lidar=True),
        ),
        need_on_same_lane=False,
        num_episode=10
    )

    # We change the ego vehicle dynamics! So the expert is not reliable anymore!
    assert 300 < ep_reward < 350, ep_reward
    # assert success_rate == 1.0, success_rate


def test_expert_without_traffic(render=False):
    ep_reward, success_rate = _evaluate(
        dict(
            num_scenarios=1,
            random_agent_model=False,
            map="CCC",
            use_render=render,
            start_seed=2,
            # debug_static_world=True,
            # debug=True,
            traffic_density=0,
            random_traffic=False,
        ),
        num_episode=10,
        has_traffic=False
    )
    # We change the ego vehicle dynamics! So the expert is not reliable anymore!
    assert success_rate == 1.0, success_rate

    assert 300 <= ep_reward <= 350, ep_reward


def test_expert_in_intersection(render=False):
    ep_reward, success_rate = _evaluate(
        dict(
            num_scenarios=1,
            random_agent_model=False,
            map="XTXTS",
            use_render=render,
            start_seed=2,
            # debug_static_world=True,
            # debug=True,
            traffic_density=0,
            random_traffic=False,
        ),
        num_episode=10,
        has_traffic=False
    )
    # We change the ego vehicle dynamics! So the expert is not reliable anymore!
    assert success_rate == 1.0, success_rate

    assert ep_reward > 400, ep_reward


if __name__ == '__main__':
    """
    LQY: I fixed a action bug in StateObservation, which may harm expert performance!
    """
    # test_expert_in_intersection(True)
    test_expert_without_traffic(True)
    # test_expert_with_traffic(use_render=True)



"""
This script takes --folder as input. It is the folder storing a batch of tfrecord file.
This script will create the output folder "processed_data" sharing the same level as `--folder`.

-- folder
-- processed_data

"""
import argparse
import copy
import os
import pickle
from collections import defaultdict

import numpy as np

from metadrive.constants import DATA_VERSION
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.utils.utils import dict_recursive_remove_array

try:
    import tensorflow as tf
except ImportError:
    pass

try:
    from waymo_open_dataset.protos import scenario_pb2
except ImportError:
    try:
        from metadrive.utils.waymo.protos import scenario_pb2  # Local files that only in PZH's computer.
    except ImportError:
        print(
            "Please install waymo_open_dataset package through metadrive dependencies: "
            "pip install waymo-open-dataset-tf-2-11-0==1.5.0"
        )

from metadrive.scenario import ScenarioDescription as SD
from metadrive.type import MetaDriveType
from metadrive.utils.waymo.utils import extract_tracks, extract_dynamic_map_states, extract_map_features, \
    compute_width
import sys

import logging

logger = logging.getLogger(__name__)

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)


def validate_sdc_track(sdc_state):
    """
    This function filters the scenario based on SDC information.

    Rule 1: Filter out if the trajectory length < 10

    Rule 2: Filter out if the whole trajectory last < 5s, assuming sampling frequency = 10Hz.
    """
    valid_array = sdc_state["valid"]
    sdc_trajectory = sdc_state["position"][valid_array, :2]
    sdc_track_length = [
        np.linalg.norm(sdc_trajectory[i] - sdc_trajectory[i + 1]) for i in range(sdc_trajectory.shape[0] - 1)
    ]
    sdc_track_length = sum(sdc_track_length)

    # Rule 1
    if sdc_track_length < 10:
        return False

    print("sdc_track_length: ", sdc_track_length)

    # Rule 2
    if valid_array.sum() < 50:
        return False

    return True


def _get_agent_summary(state_dict, id, type):
    track = state_dict["position"]
    valid_track = track[state_dict["valid"], :2]
    distance = float(sum(np.linalg.norm(valid_track[i] - valid_track[i + 1]) for i in range(valid_track.shape[0] - 1)))
    valid_length = int(sum(state_dict["valid"]))

    continuous_valid_length = 0
    for v in state_dict["valid"]:
        if v:
            continuous_valid_length += 1
        if continuous_valid_length > 0 and not v:
            break

    return {
        "type": type,
        "object_id": str(id),
        "track_length": int(len(track)),
        "distance": float(distance),
        "valid_length": int(valid_length),
        "continuous_valid_length": int(continuous_valid_length)
    }


def _get_number_summary(scenario):
    number_summary_dict = {}
    number_summary_dict["object"] = len(scenario[SD.TRACKS])
    number_summary_dict["dynamic_object_states"] = len(scenario[SD.DYNAMIC_MAP_STATES])
    number_summary_dict["map_features"] = len(scenario[SD.MAP_FEATURES])
    number_summary_dict["object_types"] = set(v["type"] for v in scenario[SD.TRACKS].values())

    object_types_counter = defaultdict(int)
    for v in scenario[SD.TRACKS].values():
        object_types_counter[v["type"]] += 1
    number_summary_dict["object_types_counter"] = dict(object_types_counter)

    # Number of different dynamic object states
    dynamic_object_states_types = set()
    dynamic_object_states_counter = defaultdict(int)
    for v in scenario[SD.DYNAMIC_MAP_STATES].values():
        for step_state in v["state"]["object_state"]:
            if step_state is None:
                continue
            dynamic_object_states_types.add(step_state)
            dynamic_object_states_counter[step_state] += 1
    number_summary_dict["dynamic_object_states_types"] = dynamic_object_states_types
    number_summary_dict["dynamic_object_states_counter"] = dict(dynamic_object_states_counter)

    return number_summary_dict


def parse_data(file_list, input_path, output_path, worker_index=None):
    scenario = scenario_pb2.Scenario()

    metadata_recorder = {}

    total_scenarios = 0

    desc = ""
    summary_file = ScenarioDescription.DATASET.SUMMARY_FILE
    if worker_index is not None:
        desc += "Worker {} ".format(worker_index)
        summary_file = "dataset_summary_worker{}.pkl".format(worker_index)

    for file_count, file in enumerate(file_list):
        file_path = os.path.join(input_path, file)
        if ("tfrecord" not in file_path) or (not os.path.isfile(file_path)):
            continue
        dataset = tf.data.TFRecordDataset(file_path, compression_type="")

        total = sum(1 for _ in dataset.as_numpy_iterator())

        for j, data in enumerate(dataset.as_numpy_iterator()):
            scenario.ParseFromString(data)

            md_scenario = SD()

            md_scenario[SD.ID] = scenario.scenario_id

            md_scenario[SD.VERSION] = DATA_VERSION

            # Please note that SDC track index is not identical to sdc_id.
            # sdc_id is a unique indicator to a track, while sdc_track_index is only the index of the sdc track
            # in the tracks datastructure.

            track_length = len(list(scenario.timestamps_seconds))

            tracks, sdc_id = extract_tracks(scenario.tracks, scenario.sdc_track_index, track_length)

            md_scenario[SD.LENGTH] = track_length

            md_scenario[SD.TRACKS] = tracks

            dynamic_states = extract_dynamic_map_states(scenario.dynamic_map_states, track_length)

            md_scenario[SD.DYNAMIC_MAP_STATES] = dynamic_states

            map_features = extract_map_features(scenario.map_features)
            md_scenario[SD.MAP_FEATURES] = map_features

            compute_width(md_scenario[SD.MAP_FEATURES])

            md_scenario[SD.METADATA] = {}
            md_scenario[SD.METADATA][SD.COORDINATE] = MetaDriveType.COORDINATE_WAYMO
            md_scenario[SD.METADATA][SD.TIMESTEP] = np.asarray(list(scenario.timestamps_seconds), dtype=np.float32)
            md_scenario[SD.METADATA][SD.METADRIVE_PROCESSED] = False
            md_scenario[SD.METADATA][SD.SDC_ID] = str(sdc_id)
            md_scenario[SD.METADATA]["dataset"] = "waymo"
            md_scenario[SD.METADATA]["scenario_id"] = scenario.scenario_id
            md_scenario[SD.METADATA]["source_file"] = str(file)
            md_scenario[SD.METADATA]["track_length"] = track_length

            # === Waymo specific data. Storing them here ===
            md_scenario[SD.METADATA]["current_time_index"] = scenario.current_time_index
            md_scenario[SD.METADATA]["sdc_track_index"] = scenario.sdc_track_index

            # obj id
            md_scenario[SD.METADATA]["objects_of_interest"] = [str(obj) for obj in scenario.objects_of_interest]

            track_index = [obj.track_index for obj in scenario.tracks_to_predict]
            track_id = [str(scenario.tracks[ind].id) for ind in track_index]
            track_difficulty = [obj.difficulty for obj in scenario.tracks_to_predict]
            track_obj_type = [tracks[id]["type"] for id in track_id]
            md_scenario[SD.METADATA]["tracks_to_predict"] = {
                id: {
                    "track_index": track_index[count],
                    "track_id": id,
                    "difficulty": track_difficulty[count],
                    "object_type": track_obj_type[count]
                }
                for count, id in enumerate(track_id)
            }

            export_file_name = SD.get_export_file_name("waymo", "v1.2" + file, scenario.scenario_id)

            summary_dict = {}
            summary_dict["sdc"] = _get_agent_summary(
                state_dict=md_scenario.get_sdc_track()["state"], id=sdc_id, type=md_scenario.get_sdc_track()["type"]
            )
            for track_id, track in md_scenario[SD.TRACKS].items():
                summary_dict[track_id] = _get_agent_summary(state_dict=track["state"], id=track_id, type=track["type"])
            md_scenario[SD.METADATA]["object_summary"] = summary_dict

            # Count some objects occurrence
            md_scenario[SD.METADATA]["number_summary"] = _get_number_summary(md_scenario)

            metadata_recorder[export_file_name] = copy.deepcopy(md_scenario[SD.METADATA])

            md_scenario = md_scenario.to_dict()

            SD.sanity_check(md_scenario, check_self_type=True)

            p = os.path.join(output_path, export_file_name)
            with open(p, "wb") as f:
                pickle.dump(md_scenario, f)

            total_scenarios += 1
            if j == total - 1:
                print(
                    f"{desc}Collected {total_scenarios} scenarios. File {file_count + 1}/{len(file_list)} has "
                    f"{total} Scenarios. The last one is saved at: {p}"
                )

    summary_file = os.path.join(output_path, summary_file)
    with open(summary_file, "wb") as file:
        pickle.dump(dict_recursive_remove_array(metadata_recorder), file)
    print("Summary is saved at: {}".format(summary_file))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", required=True, help="The data folder storing raw tfrecord from Waymo dataset.")
    parser.add_argument(
        "--output", default="processed_data", type=str, help="The data folder storing raw tfrecord from Waymo dataset."
    )
    args = parser.parse_args()

    scenario_data_path = args.input

    output_path: str = os.path.dirname(scenario_data_path)
    output_path = os.path.join(output_path, args.output)
    os.makedirs(output_path, exist_ok=True)

    raw_data_path = scenario_data_path

    # parse raw data from input path to output path,
    # there is 1000 raw data in google cloud, each of them produce about 500 pkl file
    file_list = os.listdir(raw_data_path)
    parse_data(file_list, raw_data_path, output_path)
    sys.exit()
    # file_path = AssetLoader.file_path("waymo", "processed", "0.pkl", return_raw_style=False)
    # data = read_waymo_data(file_path)
    # draw_waymo_map(data)



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import json


def get_object_properties(obj):
    text = bpy.data.texts.get(obj.name)
    if text:
        return json.loads(text.as_string() or '{}')
    else:
        return {}


def is_collision(obj):
    return obj.rigid_body is not None


def is_object_visible(obj):
    for collection in bpy.data.collections:
        if obj.name in collection.objects:
            if collection.hide_viewport:
                return False

    if obj.hide_viewport:
        return False

    return True


def set_active_object(obj):
    bpy.context.view_layer.objects.active = obj


def apply_modifiers(obj, triangulate=False, apply_scale=False):
    is_activated = False

    # if triangulate:
    #     bpy.ops.object.modifier_add(type='TRIANGULATE')
    #     obj.modifiers[-1].keep_custom_normals = True

    for mod in obj.modifiers:
        if not mod or not mod.show_viewport:
            continue

        if mod.type in ('ARMATURE', 'COLLISION'):
            continue

        if not is_activated:
            bpy.ops.object.select_all(action='DESELECT')
            obj.select_set(state=True)
            set_active_object(obj)
            is_activated = True

        try:
            bpy.ops.object.modifier_apply(modifier=mod.name)
        except Exception as e:
            print(
                'FAILED TO APPLY MODIFIER {mod_name} [{mod_type}] ON OBJECT {obj_name}'.format(
                    **{
                        'mod_name': mod.name,
                        'mod_type': mod.type,
                        'obj_name': obj.name,
                    }
                )
            )
            raise e

    if apply_scale:
        if not is_activated:
            bpy.ops.object.select_all(action='DESELECT')
            obj.select_set(state=True)
            set_active_object(obj)
            is_activated = True

        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)


def get_parent(obj, level=1):
    result = obj

    for i in range(level):
        if result is None:
            return None
        result = result.parent

    return result



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from panda3d.egg import EggMaterial

from metadrive.third_party.kitsunetsuki.base.material import get_root_node, get_from_node

SHADING_MODEL_DEFAULT = 0
SHADING_MODEL_EMISSIVE = 1
SHADING_MODEL_CLEARCOAT = 2
SHADING_MODEL_TRANSPARENT_GLASS = 3
SHADING_MODEL_SKIN = 4
SHADING_MODEL_FOLIAGE = 5
SHADING_MODEL_TRANSPARENT_EMISSIVE = 6


class MaterialMixin(object):
    def make_material(self, material):
        egg_material = EggMaterial(material.name)

        shader = None
        if material.node_tree is not None:
            output = get_root_node(material.node_tree, 'OUTPUT_MATERIAL')
            shader = None
            if output:
                shader = get_from_node(
                    material.node_tree,
                    'BSDF_PRINCIPLED',
                    to_node=output,
                    from_socket_name='BSDF',
                    to_socket_name='Surface'
                )

        if not shader:
            return egg_material

        if self._render_type == 'rp':  # RenderPipeline
            # emission = tuple(shader.inputs['Emission'].default_value)[:3]
            emission = (0, 0, 0, 0)
            # alpha = shader.inputs['Alpha'].default_value
            alpha = 1
            clearcoat = shader.inputs['Clearcoat'].default_value
            normal_strength = self.get_normal_strength(material, shader)

            if sum(emission) > 0:  # emission
                egg_material.set_base(emission + (1, ))
                egg_material.set_metallic(0)
                egg_material.set_roughness(1)
                egg_material.set_ior(1.51)

                if alpha < 1:
                    emit = [SHADING_MODEL_TRANSPARENT_EMISSIVE, normal_strength, alpha, 0]
                else:
                    emit = [SHADING_MODEL_EMISSIVE, normal_strength, 0, 0]
                egg_material.set_emit(tuple(emit))

            else:  # not emission
                egg_material.set_base((1, 1, 1, 1))
                egg_material.set_metallic(self.get_metallic(material, shader))
                egg_material.set_roughness(self.get_roughness(material, shader))
                egg_material.set_ior(shader.inputs['IOR'].default_value)

                if alpha < 1:
                    emit = [SHADING_MODEL_TRANSPARENT_GLASS, normal_strength, alpha, 0]
                elif clearcoat:
                    emit = [SHADING_MODEL_CLEARCOAT, normal_strength, 0, 0]
                else:
                    emit = [SHADING_MODEL_DEFAULT, normal_strength, 0, 0]
                egg_material.set_emit(tuple(emit))

        else:  # not RenderPipeline
            emission = (0, 0, 0, 0)
            alpha = 1

            egg_material.set_base((1, 1, 1, alpha))
            egg_material.set_metallic(self.get_metallic(material, shader))
            egg_material.set_roughness(self.get_roughness(material, shader))
            egg_material.set_ior(shader.inputs['IOR'].default_value)
            egg_material.set_emit(emission)

        return egg_material



"""

Precomputed GI using spherical harmonics.

"""

from __future__ import division, print_function

import os
import sys

os.chdir(os.path.realpath(os.path.dirname(__file__)))
sys.path.insert(0, "../../")

# Shouldn't import everything, but saves a lot of code
from panda3d.core import *
from direct.showbase.ShowBase import ShowBase

from rpcore.globals import Globals
from rpcore.render_target import RenderTarget
from rplibs.progressbar import ETA, ProgressBar, Percentage, Bar, Counter, Rate


class Application(ShowBase):
    def __init__(self):

        # Load settings
        load_prc_file_data(
            "", """
            win-size 10 10
            window-type offscreen
            win-title GI Bake
            textures-power-2 none
            gl-coordinate-system default
            sync-video #f
            support-stencil #f
            framebuffer-stencil #f
            framebuffer-multisample #f
            multisamples 0
            gl-cube-map-seamless #t
            gl-force-fbo-color #f
        """
        )

        ShowBase.__init__(self)
        Globals.load(self)
        Globals.resolution = LVecBase2i(1600, 900)
        sun_vector = Vec3(0.2, 0.5, 1.2).normalized()
        capture_resolution = 32
        sun_shadow_map_resolution = 2048
        num_probes = LVecBase3i(64)
        divisor = 128
        num_bakers = 8
        padding = 0.5

        for region in self.win.get_display_regions():
            region.set_active(False)

        print("Loading scene ...")
        # model = loader.load_model("resources/test-scene.bam")
        model = loader.load_model("scene/scene.bam")
        # model = loader.load_model("scene/LivingRoom.egg")
        model.reparent_to(render)
        model.flatten_strong()

        start_point, end_point = model.get_tight_bounds()
        start_point -= padding
        end_point += padding
        diameter = (start_point - end_point).length() * 0.5
        model_center = (start_point + end_point) * 0.5
        model_size = end_point - start_point

        print("Rendering sun shadow map ..")
        sun_shadow_cam = Camera("SunShadowCamera")
        sun_shadow_lens = OrthographicLens()
        sun_shadow_lens.set_film_size(diameter * 2, diameter * 2)
        sun_shadow_lens.set_near_far(0, 2 * diameter)
        sun_shadow_cam.set_lens(sun_shadow_lens)
        sun_shadow_cam_np = render.attach_new_node(sun_shadow_cam)
        sun_shadow_cam_np.set_pos(model_center + sun_vector * diameter)
        sun_shadow_cam_np.look_at(model_center)

        sun_shadow_target = RenderTarget()
        sun_shadow_target.size = sun_shadow_map_resolution
        sun_shadow_target.add_depth_attachment(bits=32)
        sun_shadow_target.prepare_render(sun_shadow_cam_np)

        self.render_frame()
        sun_shadow_target.active = False
        shadow_mvp = self.get_mvp(sun_shadow_cam_np)

        print("Computing first bounce ..")

        # Target to store all results
        max_probes = num_probes.x * num_probes.y * num_probes.z

        if max_probes % divisor != 0:
            print("WARNING: Bad divisor:", divisor, "for", max_probes)

        num_rows = (max_probes + divisor - 1) // divisor
        final_data = Texture("FinalProbeResult")
        final_data.setup_2d_texture(6 * divisor, num_rows, Texture.T_float, Texture.F_rgba16)
        final_data.set_clear_color(Vec4(1.0, 0.6, 0.2, 1.0))

        worker_handles = []

        store_shader = Shader.load(Shader.SL_GLSL, "resources/default.vert.glsl", "resources/copy_cubemap.frag.glsl")
        convolute_shader = Shader.load(Shader.SL_GLSL, "resources/default.vert.glsl", "resources/convolute.frag.glsl")

        for worked_id in range(num_bakers):
            probe_position = Vec3(0, 0, 4)
            capture_target = RenderTarget()
            capture_target.size = capture_resolution * 6, capture_resolution
            capture_target.add_depth_attachment(bits=16)
            capture_target.add_color_attachment(bits=16, alpha=True)
            capture_target.prepare_render(None)

            # Remove all unused display regions
            internal_buffer = capture_target.internal_buffer
            internal_buffer.remove_all_display_regions()
            internal_buffer.disable_clears()
            internal_buffer.get_overlay_display_region().disable_clears()

            # Setup the cubemap capture rig
            directions = (Vec3(1, 0, 0), Vec3(-1, 0, 0), Vec3(0, 1, 0), Vec3(0, -1, 0), Vec3(0, 0, 1), Vec3(0, 0, -1))
            capture_regions = []
            capture_cams = []
            capture_rig = render.attach_new_node("CaptureRig")
            capture_rig.set_pos(probe_position)

            # Prepare the display regions
            for i in range(6):
                region = capture_target.internal_buffer.make_display_region(i / 6, i / 6 + 1 / 6, 0, 1)
                region.set_sort(25 + i)
                region.set_active(True)
                region.disable_clears()

                # Set the correct clears
                region.set_clear_depth_active(True)
                region.set_clear_depth(1.0)
                region.set_clear_color_active(True)
                region.set_clear_color(Vec4(0.0, 0.0, 0.0, 0.0))

                lens = PerspectiveLens()
                lens.set_fov(90)
                lens.set_near_far(0.05, 2 * diameter)
                camera = Camera("CaptureCam-" + str(i), lens)
                camera_np = capture_rig.attach_new_node(camera)
                camera_np.look_at(camera_np, directions[i])
                region.set_camera(camera_np)
                capture_regions.append(region)
                capture_cams.append(camera_np)

            capture_cams[0].set_r(90)
            capture_cams[1].set_r(-90)
            capture_cams[3].set_r(180)
            capture_cams[5].set_r(180)

            destination_cubemap = Texture("TemporaryCubemap")
            destination_cubemap.setup_cube_map(capture_resolution, Texture.T_float, Texture.F_rgba16)

            # Target to convert the FBO to a cubemap
            target_store_cubemap = RenderTarget()
            target_store_cubemap.size = capture_resolution * 6, capture_resolution
            target_store_cubemap.prepare_buffer()
            target_store_cubemap.set_shader_inputs(SourceTex=capture_target.color_tex, DestTex=destination_cubemap)

            target_store_cubemap.shader = store_shader

            # Target to filter the data
            store_pta = PTALVecBase2i.empty_array(1)
            target_convolute = RenderTarget()
            target_convolute.size = 6, 1
            # target_convolute.add_color_attachment(bits=16)
            target_convolute.prepare_buffer()
            target_convolute.set_shader_inputs(SourceTex=destination_cubemap, DestTex=final_data, storeCoord=store_pta)
            target_convolute.shader = convolute_shader

            # Set initial shader
            shader = Shader.load(Shader.SL_GLSL, "resources/first-bounce.vert.glsl", "resources/first-bounce.frag.glsl")
            render.set_shader(shader)
            render.set_shader_inputs(ShadowMap=sun_shadow_target.depth_tex, shadowMVP=shadow_mvp, sunVector=sun_vector)

            worker_handles.append((capture_rig, store_pta))

        print("Preparing to render", max_probes, "probes ..")
        widgets = [Counter(), "  ", Bar(), "  ", Percentage(), "  ", ETA(), " ", Rate()]
        progressbar = ProgressBar(widgets=widgets, maxval=max_probes).start()
        progressbar.update(0)

        work_queue = []

        for z_pos in range(num_probes.z):
            for y_pos in range(num_probes.y):
                for x_pos in range(num_probes.x):
                    index = x_pos + y_pos * num_probes.x + z_pos * num_probes.y * num_probes.x
                    #     print("Baking", index, "out of", max_probes)
                    offs_x = start_point.x + x_pos / (num_probes.x + 0.5) * model_size.x
                    offs_y = start_point.y + y_pos / (num_probes.y + 0.5) * model_size.y
                    offs_z = start_point.z + z_pos / (num_probes.z + 0.5) * model_size.z

                    store_x = index % divisor
                    store_y = index // divisor

                    work_queue.append((Vec3(offs_x, offs_y, offs_z), LVecBase2i(store_x, store_y)))

        for i, (pos, store) in enumerate(work_queue):
            worker_handles[i % num_bakers][0].set_pos(pos)
            worker_handles[i % num_bakers][1][0] = store
            if i % num_bakers == num_bakers - 1:
                self.render_frame()
                progressbar.update(i)

        progressbar.finish()
        self.render_frame()

        print("Writing out data ..")
        self.graphicsEngine.extract_texture_data(final_data, self.win.gsg)
        final_data.write("raw-bake.png")

        print("Writing out configuration")
        with open("_bake_params.py", "w") as handle:
            handle.write("#Autogenerated\n")
            handle.write("from panda3d.core import LPoint3f, LVecBase3i, LVector3f\n")
            handle.write("BAKE_MESH_START = " + str(start_point) + "\n")
            handle.write("BAKE_MESH_END = " + str(end_point) + "\n")
            handle.write("BAKE_MESH_PROBECOUNT = " + str(num_probes) + "\n")
            handle.write("BAKE_DIVISOR = " + str(divisor) + "\n")
            handle.write("BAKE_SUN_VECTOR = " + str(sun_vector) + "\n")

        with open("_bake_params.glsl", "w") as handle:
            handle.write("// Autogenerated\n")
            handle.write("#define LPoint3f vec3\n")
            handle.write("#define LVector3f vec3\n")
            handle.write("#define LVecBase3i ivec3\n")
            handle.write("const vec3 bake_mesh_start = " + str(start_point) + ";\n")
            handle.write("const vec3 bake_mesh_end = " + str(end_point) + ";\n")
            handle.write("const ivec3 bake_mesh_probecount = " + str(num_probes) + ";\n")
            handle.write("const int bake_divisor = " + str(divisor) + ";\n")
            handle.write("const vec3 sun_vector = " + str(sun_vector) + ";\n")

    def render_frame(self):
        """ Convenience function to render a frame """
        self.graphicsEngine.render_frame()

    def get_mvp(self, cam_node):
        """ Computes the view-projection matrix of a camera """
        return render.get_transform(cam_node).get_mat() * cam_node.node().get_lens().get_projection_mat()


Application()



from metadrive.component.static_object.traffic_object import TrafficCone, TrafficWarning
from metadrive.component.vehicle.vehicle_type import LVehicle
from metadrive.constants import TerminationState, DEFAULT_AGENT
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.manager.object_manager import TrafficObjectManager


class ComplexObjectManager(TrafficObjectManager):
    def reset(self):
        ret = super(ComplexObjectManager, self).reset()
        lane = self.engine.current_map.road_network.graph[">>>"]["1C0_0_"][0]
        breakdown_vehicle = self.engine.object_manager.spawn_object(
            self.engine.traffic_manager.random_vehicle_type(),
            vehicle_config={
                "spawn_lane_index": lane.index,
                "spawn_longitude": 30
            }
        )
        self.engine.object_manager.accident_lanes.append(lane)
        lane_ = self.engine.current_map.road_network.graph[">>>"]["1C0_0_"][1]
        breakdown_vehicle = self.engine.object_manager.spawn_object(
            LVehicle, vehicle_config={
                "spawn_lane_index": lane_.index,
                "spawn_longitude": 30
            }
        )
        self.engine.object_manager.accident_lanes.append(lane_)
        lane = lane
        longitude = 22
        lateral = 0
        alert = self.engine.object_manager.spawn_object(
            TrafficWarning,
            lane=lane,
            position=lane.position(longitude, lateral),
            heading_theta=lane.heading_theta_at(longitude),
            pbr_model=False
        )

        # part 1
        lane = self.engine.current_map.road_network.graph["1C0_1_"]["2S0_0_"][2]
        pos = [
            (-20, lane.width / 3), (-15.6, lane.width / 4), (-12.1, 0), (-8.7, -lane.width / 4),
            (-4.2, -lane.width / 2), (-0.7, -lane.width * 3 / 4), (4.1, -lane.width), (7.3, -lane.width),
            (11.5, -lane.width), (15.5, -lane.width), (20.0, -lane.width), (23.2, -lane.width), (29.1, -lane.width),
            (32.9, -lane.width / 2), (37.0, 0), (40.0, lane.width / 2)
        ]

        for p in pos:
            longitude = p[0]
            lateral = p[1] + lane.width / 2
            cone = self.engine.object_manager.spawn_object(
                TrafficCone,
                lane=lane,
                position=lane.position(longitude, lateral),
                heading_theta=lane.heading_theta_at(longitude)
            )
        self.engine.object_manager.accident_lanes.append(lane)
        from metadrive.component.vehicle.vehicle_type import SVehicle, XLVehicle
        v_pos = [8, 14]
        v_type = [SVehicle, XLVehicle]
        for v_long, v_t in zip(v_pos, v_type):
            v = self.engine.object_manager.spawn_object(
                v_t, vehicle_config={
                    "spawn_lane_index": lane.index,
                    "spawn_longitude": v_long
                }
            )

        # part 2
        lane = self.engine.current_map.road_network.graph["3R0_0_"]["3R0_1_"][0]
        self.engine.object_manager.accident_lanes.append(lane)
        pos = [
            (-20, lane.width / 3), (-15.6, lane.width / 4), (-12.1, 0), (-8.7, -lane.width / 4),
            (-4.2, -lane.width / 2), (-0.7, -lane.width * 3 / 4), (4.1, -lane.width), (7.3, -lane.width),
            (11.5, -lane.width), (15.5, -lane.width), (20.0, -lane.width), (23.2, -lane.width), (29.1, -lane.width),
            (32.9, -lane.width / 2), (37.0, 0), (40.0, lane.width / 2)
        ]

        for p in pos:
            p_ = (p[0] + 5, -p[1])
            longitude = p_[0]
            lateral = -p[1] - lane.width / 2
            cone = self.engine.object_manager.spawn_object(
                TrafficCone,
                lane=lane,
                position=lane.position(longitude, lateral),
                heading_theta=lane.heading_theta_at(longitude)
            )

        v_pos = [14, 19]
        for v_long in v_pos:
            v = self.engine.object_manager.spawn_object(
                self.engine.traffic_manager.random_vehicle_type(),
                vehicle_config={
                    "spawn_lane_index": lane.index,
                    "spawn_longitude": v_long
                }
            )
        longitude = -35
        lateral = 0
        alert = self.engine.object_manager.spawn_object(
            TrafficWarning,
            lane=lane,
            position=lane.position(longitude, lateral),
            heading_theta=lane.heading_theta_at(longitude)
        )
        longitude = -60
        lateral = 0
        alert = self.engine.object_manager.spawn_object(
            TrafficWarning,
            lane=lane,
            position=lane.position(longitude, lateral),
            heading_theta=lane.heading_theta_at(longitude)
        )

        # part 3
        lane = self.engine.current_map.road_network.graph["4C0_0_"]["4C0_1_"][2]
        self.engine.object_manager.accident_lanes.append(lane)
        pos = [
            (-12.1, 0), (-8.7, -lane.width / 4), (-4.2, -lane.width / 2), (-0.7, -lane.width * 3 / 4),
            (4.1, -lane.width), (7.3, -lane.width), (11.5, -lane.width), (15.5, -lane.width), (20.0, -lane.width),
            (23.2, -lane.width), (29.1, -lane.width), (32.9, -lane.width / 2), (37.0, 0), (40.0, lane.width / 2)
        ]

        for p in pos:
            p_ = (p[0] + 5, p[1] * 3.5 / 3)
            longitude = p_[0]
            lateral = p[1] + lane.width / 2
            cone = self.engine.object_manager.spawn_object(
                TrafficCone,
                lane=lane,
                position=lane.position(longitude, lateral),
                heading_theta=lane.heading_theta_at(longitude)
            )

        v_pos = [14, 19]
        for v_long in v_pos:
            v = self.engine.object_manager.spawn_object(
                self.engine.traffic_manager.random_vehicle_type(),
                vehicle_config={
                    "spawn_lane_index": lane.index,
                    "spawn_longitude": v_long
                }
            )

        # part 4
        lane = self.engine.current_map.road_network.graph["4C0_1_"]["5R0_0_"][0]
        self.engine.object_manager.accident_lanes.append(lane)
        # pos = [(-12, lane.width / 4), (-8.1, 0), (-4, -lane.width / 4), (-0.1, -lane.width / 2), (4, -lane.width)]
        #
        # for p in pos:
        #     p_ = (p[0] + 60, -p[1] * 3.5 / 3)
        #     cone = self.engine.object_manager.spawn_object(TrafficCone, lane=lane, longitude=p_[0], lateral=p_[1])

        return ret


class ComplexEnv(SafeMetaDriveEnv):
    """
    now for test use and demo use only
    """
    def default_config(self):
        config = super(ComplexEnv, self).default_config()
        config.update(
            {
                "num_scenarios": 1,
                "traffic_density": 0.05,
                "start_seed": 5,
                "accident_prob": 0.0,
                # "traffic_mode":"respawn",
                "debug_physics_world": False,
                "debug": False,
                "map": "CSRCR"
            }
        )
        return config

    def setup_engine(self):
        super(ComplexEnv, self).setup_engine()
        self.engine.register_manager("object_manager", ComplexObjectManager())


def test_object_collision_detection(render=False):
    env = ComplexEnv(
        {
            # "manual_control": True,
            "traffic_density": 0.0,
            "use_render": render,
            "crash_object_cost": 100,
            "crash_object_done": True,
            "debug": False,
            "vehicle_config": {
                "show_lidar": True
            }
        }
    )
    try:
        o, _ = env.reset()
        lane_index = (">>", ">>>", 0)
        lane = env.current_map.road_network.get_lane(lane_index)
        longitude = 22
        lateral = 0
        alert = env.engine.object_manager.spawn_object(
            TrafficWarning,
            lane=env.current_map.road_network.get_lane(lane_index),
            position=lane.position(longitude, lateral),
            heading_theta=lane.heading_theta_at(longitude)
        )
        lane_index = (">>", ">>>", 2)

        longitude = 22
        lateral = 0
        lane = env.current_map.road_network.get_lane(lane_index)
        alert = env.engine.object_manager.spawn_object(
            TrafficCone,
            lane=env.current_map.road_network.get_lane(lane_index),
            position=lane.position(longitude, lateral),
            heading_theta=lane.heading_theta_at(longitude)
        )
        crash_obj = False
        detect_obj = False
        for i in range(1, 100000 if render else 2000):
            o, r, tm, tc, info = env.step([0, 1])
            for obj in env.observations[DEFAULT_AGENT].detected_objects:
                if isinstance(obj, TrafficCone):
                    detect_obj = True
            if render:
                env.render()
            if info["cost"] == 100 and info[TerminationState.CRASH_OBJECT]:
                crash_obj = True
                break
        assert crash_obj and detect_obj, "Can not crash with object!"
    finally:
        env.close()


if __name__ == "__main__":
    test_object_collision_detection(render=False)



from metadrive import MetaDriveEnv
from metadrive.engine.asset_loader import initialize_asset_loader, AssetLoader
from metadrive.engine.core.engine_core import EngineCore


def test_asset_loader():
    default_config = MetaDriveEnv.default_config()
    world = EngineCore(global_config=default_config)
    try:
        world.clear_world()
        initialize_asset_loader(world)
        # print(AssetLoader.asset_path)
        # print(AssetLoader.file_path("aaa"))
    # # print(AssetLoader.get_loader())
    finally:
        world.close_world()


if __name__ == '__main__':
    test_asset_loader()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import print_function

from panda3d.core import Vec3, Mat3


def color_from_temperature(temperature):
    # Thanks to rdb for this conversion script
    mm = 1000.0 / temperature  # pylint: disable=invalid-name
    mm2 = mm**2
    mm3 = mm2 * mm
    x, y = 0, 0

    if temperature < 4000:
        x = -0.2661239 * mm3 - 0.2343580 * mm2 + 0.8776956 * mm + 0.179910
    else:
        x = -3.0258469 * mm3 + 2.1070379 * mm2 + 0.2226347 * mm + 0.240390

    x2 = x**2  # pylint: disable=invalid-name
    x3 = x2 * x  # pylint: disable=invalid-name
    if temperature < 2222:
        y = -1.1063814 * x3 - 1.34811020 * x2 + 2.18555832 * x - 0.20219683
    elif temperature < 4000:
        y = -0.9549476 * x3 - 1.37418593 * x2 + 2.09137015 * x - 0.16748867
    else:
        y = 3.0817580 * x3 - 5.87338670 * x2 + 3.75112997 * x - 0.37001483

    # xyY to XYZ, assuming Y=1.
    xyz = Vec3(x / y, 1, (1 - x - y) / y)

    # Convert XYZ to linearized sRGB.
    xyz_to_rgb = Mat3(3.2406, -0.9689, 0.0557, -1.5372, 1.8758, -0.2050, -0.4986, 0.0415, 1.0570)

    return xyz_to_rgb.xform(xyz)


__all__ = ["RPLight"]


class RPLight(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """

    LT_empty = 0
    LT_point_light = 1
    LT_spot_light = 2

    def __init__(self, light_type):
        self._light_type = light_type
        self._needs_update = False
        self._casts_shadows = False
        self._slot = -1
        self._position = Vec3(0)
        self._color = Vec3(1)
        self._ies_profile = -1
        self._source_resolution = 512
        self._near_plane = 0.5
        self._energy = 20
        self._shadow_sources = []

    def get_num_shadow_sources(self):
        return len(self._shadow_sources)

    def get_shadow_source(self, index):
        return self._shadow_sources[index]

    def clear_shadow_sources(self):
        self._shadow_sources = []

    def set_needs_update(self, flag):
        self._needs_update = flag

    def get_needs_update(self):
        return self._needs_update

    def has_slot(self):
        return self._slot >= 0

    def get_slot(self):
        return self._slot

    def remove_slot(self):
        self._slot = -1

    def assign_slot(self, slot):
        self._slot = slot

    def invalidate_shadows(self):
        for source in self._shadow_sources:
            source.set_needs_update(True)

    def set_pos(self, *args):
        self._position = Vec3(*args)
        self.set_needs_update(True)
        self.invalidate_shadows()

    def get_pos(self):
        return self._position

    pos = property(get_pos, set_pos)

    def set_color(self, *args):
        self._color = Vec3(*args)
        self._color /= (0.2126 * self._color.x + 0.7152 * self._color.y + 0.0722 * self._color.z)
        self.set_needs_update(True)

    def set_color_from_temperature(self, temperature):
        self.set_color(color_from_temperature(temperature))

    def get_color(self):
        return self._color

    color = property(get_color, set_color)

    def set_energy(self, energy):
        self._energy = energy

    def get_energy(self):
        return self._energy

    energy = property(get_energy, set_energy)

    def get_light_type(self):
        return self._light_type

    light_type = property(get_light_type)

    def set_casts_shadows(self, flag):
        if self.has_slot():
            print("Light is already attached!")
            return
        self._casts_shadows = flag

    def get_casts_shadows(self):
        return self._casts_shadows

    casts_shadows = property(get_casts_shadows, set_casts_shadows)

    def set_shadow_map_resolution(self, resolution):
        self._source_resolution = resolution
        self.invalidate_shadows()

    def get_shadow_map_resolution(self):
        return self._source_resolution

    shadow_map_resolution = property(get_shadow_map_resolution, set_shadow_map_resolution)

    def set_ies_profile(self, profile):
        self._ies_profile = profile
        self.set_needs_update(True)

    def get_ies_profile(self):
        return self._ies_profile

    def has_ies_profile(self):
        return self._ies_profile >= 0

    def clear_ies_profile(self):
        self.set_ies_profile(-1)

    ies_profile = property(get_ies_profile, set_ies_profile)

    def set_near_plane(self, near_plane):
        self._near_plane = near_plane
        self.invalidate_shadows()

    def get_near_plane(self):
        return self._near_plane

    near_plane = property(get_near_plane, set_near_plane)

    def write_to_command(self, cmd):
        cmd.push_int(self._light_type)
        cmd.push_int(self._ies_profile)

        if self._casts_shadows:
            cmd.push_int(self._shadow_sources[0].get_slot())
        else:
            cmd.push_int(-1)

        cmd.push_vec3(self._position)
        cmd.push_vec3(self._color * self._energy / 100.0)



import logging
from typing import List, TYPE_CHECKING, Tuple, Union

import math
import numpy as np
from panda3d.bullet import BulletBoxShape, BulletCylinderShape, ZUp
from panda3d.core import TransformState
from panda3d.core import Vec3

from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.lane.pg_lane import PGLane
from metadrive.constants import CollisionGroup
from metadrive.constants import Decoration, MetaDriveType
from metadrive.engine.core.engine_core import EngineCore
from metadrive.engine.physics_node import BaseRigidBodyNode, BaseGhostBodyNode
from metadrive.utils.coordinates_shift import panda_heading
from metadrive.utils.coordinates_shift import panda_vector
from metadrive.utils.interpolating_line import InterpolatingLine
from metadrive.utils.math import get_points_bounding_box, norm
from metadrive.utils.utils import get_object_from_node

if TYPE_CHECKING:
    from metadrive.component.pgblock.pg_block import PGBlockSocket
    from metadrive.component.road_network.node_road_network import NodeRoadNetwork


def block_socket_merge(
    socket_1: "PGBlockSocket", socket_2: "PGBlockSocket", global_network: "NodeRoadNetwork", positive_merge: False
):
    global_network.graph[socket_1.positive_road.start_node][socket_2.negative_road.start_node] = \
        global_network.graph[socket_1.positive_road.start_node].pop(socket_1.positive_road.end_node)

    global_network.graph[socket_2.positive_road.start_node][socket_1.negative_road.start_node] = \
        global_network.graph[socket_2.positive_road.start_node].pop(socket_2.positive_road.end_node)


def check_lane_on_road(
    road_network: "NodeRoadNetwork",
    lane,
    positive: float = 0,
    ignored=None,
    ignore_intersection_checking=None
) -> bool:
    """
    Calculate if the new lane intersects with other lanes in current road network
    The return Value is True when cross
    Note: the decoration road will be ignored in default
    """
    assert ignore_intersection_checking is not None
    if ignore_intersection_checking:
        return True
    graph = road_network.graph
    for _from, to_dict in graph.items():
        for _to, lanes in to_dict.items():
            if ignored and (_from, _to) == ignored:
                continue
            if (_from, _to) == (Decoration.start, Decoration.end):
                continue
            if len(lanes) == 0:
                continue
            x_max_1, x_min_1, y_max_1, y_min_1 = get_lanes_bounding_box(lanes)
            x_max_2, x_min_2, y_max_2, y_min_2 = get_lanes_bounding_box([lane])
            if x_min_1 > x_max_2 or x_min_2 > x_max_1 or y_min_1 > y_max_2 or y_min_2 > y_max_1:
                continue
            for _id, l in enumerate(lanes):
                for i in range(1, int(lane.length), 1):
                    sample_point = lane.position(i, positive * lane.width_at(i) / 2.0)
                    longitudinal, lateral = l.local_coordinates(sample_point)
                    is_on = math.fabs(lateral) <= l.width_at(longitudinal) / 2.0 and 0 <= longitudinal <= l.length
                    if is_on:
                        return True
    return False


def get_lanes_bounding_box(lanes, extra_lateral=3) -> Tuple:
    """
    Return (x_max, x_min, y_max, y_min) as bounding box of this road
    :param lanes: Lanes in this road
    :param extra_lateral: extra width in lateral direction, usually sidewalk width
    :return: x_max, x_min, y_max, y_min
    """
    if isinstance(lanes[0], PGLane):
        line_points = get_curve_contour(lanes, extra_lateral) if isinstance(lanes[0], CircularLane) \
            else get_straight_contour(lanes, extra_lateral)
    else:
        line_points = get_interpolating_lane_countour(lanes)
    return get_points_bounding_box(line_points)


def get_interpolating_lane_countour(lanes):
    assert isinstance(lanes[0], InterpolatingLine)
    ret = []
    for lane in lanes:
        for seg in lane.segment_property:
            ret.append(seg["start_point"])
        ret.append(seg["end_point"])
    return ret


def get_straight_contour(lanes, extra_lateral) -> List:
    """
    Get several points as bounding box of this road
    :param lanes: lanes contained in road
    :param extra_lateral: extra length in lateral direction, usually sidewalk
    :return: points
    :param lanes:
    :return:
    """
    ret = []
    for lane, dir in [(lanes[0], -1), (lanes[-1], 1)]:
        ret.append(lane.position(0.1, dir * (lane.width / 2.0 + extra_lateral)))
        ret.append(lane.position(lane.length - 0.1, dir * (lane.width / 2.0 + extra_lateral)))
    return ret


def get_curve_contour(lanes, extra_lateral) -> List:
    """
    Get several points as bounding box of this road
    :param lanes: lanes contained in road
    :param extra_lateral: extra length in lateral direction, usually sidewalk
    :return: points
    """
    points = []
    for lane, lateral_dir in [(lanes[0], -1), (lanes[-1], 1)]:
        pi_2 = (np.pi / 2.0)
        points += [
            lane.position(0.1, lateral_dir * (lane.width / 2.0 + extra_lateral)),
            lane.position(lane.length - 0.1, lateral_dir * (lane.width / 2.0 + extra_lateral))
        ]
        start_phase = (lane.start_phase // pi_2) * pi_2
        start_phase += pi_2 if lane.is_clockwise() else 0
        for phi_index in range(4):
            phi = start_phase + phi_index * pi_2 * lane.direction
            if lane.direction * phi > lane.direction * lane.end_phase:
                break
            point = lane.center + (lane.radius - lateral_dir * (lane.width / 2.0 + extra_lateral) *
                                   lane.direction) * np.array([math.cos(phi), math.sin(phi)])
            points.append(point)
    return points


def get_all_lanes(roadnet: "NodeRoadNetwork"):
    graph = roadnet.graph
    res = []
    for from_, to_dict in graph.items():
        for _to, lanes in to_dict.items():
            for l in lanes:
                res.append(l)
    return res


def ray_localization(
    heading: tuple,
    position: tuple,
    engine: EngineCore,
    use_heading_filter=True,
    return_on_lane=False,
) -> Union[List[Tuple], Tuple]:
    """
    Get the index of the lane closest to a physx_world position.
    Only used when smoething is on lane ! Otherwise fall back to use get_closest_lane()
    :param heading: heading to help filter lanes
    :param position: a physx_world position [m].
    :param engine: BaseEngine class
    :param return_all_result: return a list instead of the lane with min L1 distance
    :return: list(closest lane) or closest lane.
    """
    on_lane = False

    if len(position) == 3:
        position = position[:2]
    assert len(position) == 2
    assert len(heading) == 2

    results = engine.physics_world.static_world.rayTestAll(panda_vector(position, 1.0), panda_vector(position, -1.0))
    lane_index_dist = []
    if results.hasHits():
        for res in results.getHits():
            if res.getNode().getName() == MetaDriveType.LANE_SURFACE_STREET:
                on_lane = True
                lane = get_object_from_node(res.getNode())
                long, _ = lane.local_coordinates(position)
                lane_heading = lane.heading_theta_at(long)

                # dir = np.array([math.cos(lane_heading), math.sin(lane_heading)])
                # dot_result = dir.dot(heading)

                dot_result = math.cos(lane_heading) * heading[0] + math.sin(lane_heading) * heading[1]
                cosangle = dot_result / (
                    norm(math.cos(lane_heading), math.sin(lane_heading)) * norm(heading[0], heading[1])
                )

                if use_heading_filter:
                    if cosangle > 0:
                        lane_index_dist.append((lane, lane.index, lane.distance(position)))
                else:
                    lane_index_dist.append((lane, lane.index, lane.distance(position)))
    # default return all result
    ret = []
    if len(lane_index_dist) > 0:
        ret = sorted(lane_index_dist, key=lambda k: k[2])

    # sorted(ret, key=lambda k: k[2]) what a stupid bug. sorted is not an inplace operation
    return (ret, on_lane) if return_on_lane else ret
    # else:
    #     if len(lane_index_dist) > 0:
    #         ret_index = np.argmin([d for _, _, d in lane_index_dist])
    #         lane, index, dist = lane_index_dist[ret_index]
    #     else:
    #         lane, index, dist = None, None, None
    #     return (lane, index) if not return_on_lane else (lane, index, on_lane)


def rect_region_detection(
    engine: EngineCore,
    position: Tuple,
    heading: float,
    heading_direction_length: float,
    side_direction_width: float,
    detection_group: int,
    height=10,
    in_static_world=False
):
    """

     ----------------------------------
     |               *                |  --->>>
     ----------------------------------
     * position
     --->>> heading direction
     ------ longitude length
     | lateral width

     **CAUTION**: position is the middle point of longitude edge

    :param engine: BaseEngine class
    :param position: position in MetaDrive
    :param heading: heading in MetaDrive [degree]
    :param heading_direction_length: rect length in heading direction
    :param side_direction_width: rect width in side direction
    :param detection_group: which group to detect
    :param height: the detect will be executed from this height to 0
    :param in_static_world: execute detection in static world
    :return: detection result
    """
    region_detect_start = panda_vector(position, z=height)
    region_detect_end = panda_vector(position, z=-1)
    tsFrom = TransformState.makePosHpr(region_detect_start, Vec3(panda_heading(heading), 0, 0))
    tsTo = TransformState.makePosHpr(region_detect_end, Vec3(panda_heading(heading), 0, 0))

    shape = BulletBoxShape(Vec3(heading_direction_length / 2, side_direction_width / 2, 1))
    penetration = 0.0

    physics_world = engine.physics_world.dynamic_world if not in_static_world else engine.physics_world.static_world

    result = physics_world.sweep_test_closest(shape, tsFrom, tsTo, detection_group, penetration)
    return result


def circle_region_detection(
    engine: EngineCore, position: Tuple, radius: float, detection_group: int, height=10, in_static_world=False
):
    """
    :param engine: BaseEngine class
    :param position: position in MetaDrive
    :param radius: radius of the region to be detected
    :param detection_group: which group to detect
    :param height: the detect will be executed from this height to 0
    :param in_static_world: execute detection in static world
    :return: detection result
    """
    region_detect_start = panda_vector(position, z=height)
    region_detect_end = panda_vector(position, z=-1)
    tsFrom = TransformState.makePos(region_detect_start)
    tsTo = TransformState.makePos(region_detect_end)

    shape = BulletCylinderShape(radius, 5, ZUp)
    penetration = 0.0

    physics_world = engine.physics_world.dynamic_world if not in_static_world else engine.physics_world.static_world

    result = physics_world.sweep_test_closest(shape, tsFrom, tsTo, detection_group, penetration)
    return result


def generate_static_box_physics_body(
    heading_length: float,
    side_width: float,
    height=10,
    ghost_node=False,
    object_id=None,
    type_name=MetaDriveType.INVISIBLE_WALL,
    collision_group=CollisionGroup.InvisibleWall
):
    """
    Add an invisible physics wall to physics world
    You can add some building models to the same location, add make it be detected by lidar
    ----------------------------------
    |               *                |  --->>>
    ----------------------------------
    * position
    --->>> heading direction
    ------ longitude length
    | lateral width

    **CAUTION**: position is the middle point of longitude edge
    :param heading_length: rect length in heading direction
    :param side_width: rect width in side direction
    :param height: the detect will be executed from this height to 0
    :param object_id: name of this invisible wall
    :param ghost_node: need physics reaction or not
    :param type_name: default invisible wall
    :param collision_group: control the collision of this static wall and other elements
    :return node_path
    """
    shape = BulletBoxShape(Vec3(heading_length / 2, side_width / 2, height))
    body_node = BaseRigidBodyNode(object_id, type_name) if not ghost_node else BaseGhostBodyNode(object_id, type_name)
    body_node.setActive(False)
    body_node.setKinematic(False)
    body_node.setStatic(True)
    body_node.addShape(shape)
    body_node.setIntoCollideMask(collision_group)
    return body_node


def is_same_lane_index(lane_index_1, lane_index_2):
    return all([lane_index_1[i] == lane_index_2[i] for i in range(3)])


def is_following_lane_index(current_lane_index, next_lane_index):
    return True if current_lane_index[1] == next_lane_index[0] and current_lane_index[-1] == next_lane_index[
        -1] else False



from metadrive.envs.marl_envs.marl_parking_lot import MultiAgentParkingLotEnv


def vis_ma_parking_lot_env():
    import cv2
    env = MultiAgentParkingLotEnv(
        {
            "use_render": False,
            "image_observation": True,
            # it is a switch telling metadrive to use rgb as observation
            "rgb_clip": True,  # clip rgb to range(0,1) instead of (0, 255)
            "delay_done": 0,
            "num_agents": 4,
            "vehicle_config": {
                "stack_size": 5,
                "rgb_camera": (800, 600),
                "lidar": {
                    "num_others": 8
                }
            }
        }
    )
    env.reset()
    o, r, tm, tc, i = env.step(env.action_space.sample())
    for i in range(4):
        cv2.imshow('img', o["agent{}".format(i)]["image"][..., -1])
        cv2.waitKey(0)


if __name__ == '__main__':
    vis_ma_parking_lot_env()



import numpy as np

from metadrive.envs.metadrive_env import MetaDriveEnv


def test_coordinates_shift():
    try:
        env = MetaDriveEnv(
            {
                "num_scenarios": 100,
                "traffic_density": .0,
                "traffic_mode": "trigger",
                "start_seed": 22,
                # "manual_control": True,
                # "use_render": True,
                "decision_repeat": 5,
                "rgb_clip": True,
                "pstats": True,
                # "discrete_action": True,
                "map": "SSSSSS",
            }
        )
        env.reset()
        env.vehicle.set_velocity([1, 0], 10)
        # print(env.vehicle.speed)
        pos = [(x, y) for x in [-10, 0, 10] for y in [-20, 0, 20]] * 10
        p = pos.pop()
        for s in range(1, 100000):
            o, r, tm, tc, info = env.step([1, 0.3])
            if s % 10 == 0:
                if len(pos) == 0:
                    break
                p = pos.pop()
            p = np.asarray(p)
            heading, side = env.vehicle.convert_to_local_coordinates(p, env.vehicle.position)
            recover_pos = env.vehicle.convert_to_world_coordinates([heading, side], env.vehicle.position)
            if abs(recover_pos[0] - p[0]) + abs(recover_pos[1] - p[1]) > 0.1:
                raise ValueError("vehicle coordinates convert error!")
    finally:
        env.close()


if __name__ == "__main__":
    test_coordinates_shift()



import os
import pathlib
import sys
from metadrive.pull_asset import pull_asset
from metadrive.engine.logger import get_logger
from metadrive.utils.utils import is_win
from metadrive.version import VERSION


class AssetLoader:
    """
    Load model for each element when render is needed.
    """
    logger = get_logger("Asset")
    loader = None
    asset_path = pathlib.PurePosixPath(__file__).parent.parent.joinpath("assets") if not is_win(
    ) else pathlib.Path(__file__).resolve().parent.parent.joinpath("assets")

    @staticmethod
    def init_loader(engine):
        """
        Due to the feature of Panda3d, keep reference of loader in static variable
        """
        if engine.win is None:
            AssetLoader.logger.debug("Physics world mode")
            return
        AssetLoader.logger.debug("Onscreen/Offscreen mode, Render/Load Elements")
        AssetLoader.loader = engine.loader

    @property
    def asset_version(self):
        return asset_version()

    @classmethod
    def get_loader(cls):
        assert AssetLoader.loader, "Please initialize AssetLoader before getting it!"
        return cls.loader

    @staticmethod
    def windows_style2unix_style(win_path):
        path = win_path.as_posix()
        panda_path = "/" + path[0].lower() + path[2:]
        return panda_path

    @staticmethod
    def file_path(*path_string, return_raw_style=True):
        """
        Usage is the same as path.join(dir_1,dir_2,file_name)
        :param path_string: a tuple
        :param return_raw_style: it will not return raw style and not do any style converting
        :return: file path used to load asset
        """
        path = AssetLoader.asset_path.joinpath(*path_string)
        return AssetLoader.windows_style2unix_style(
            path
        ) if sys.platform.startswith("win") and return_raw_style else str(path)

    @classmethod
    def load_model(cls, file_path):
        """
        A quick load method
        :param file_path: path in string, usually use the return value of AssetLoader.file_path()
        :return: model node path
        """
        assert cls.loader is not None
        return cls.loader.loadModel(file_path)

    @classmethod
    def initialized(cls):
        return cls.loader is not None


def initialize_asset_loader(engine):
    # load model file in utf-8
    os.environ["PYTHONUTF8"] = "on"
    if AssetLoader.initialized():
        AssetLoader.logger.warning(
            "AssetLoader is initialize to root path: {}! But you are initializing again!".format(
                AssetLoader.asset_path
            )
        )
        return
    AssetLoader.init_loader(engine)


def close_asset_loader():
    cls = AssetLoader
    cls.loader = None


def randomize_cover():
    background_folder_name = "background"
    files = os.listdir(AssetLoader.asset_path.joinpath(background_folder_name))
    files = [f for f in files if f.startswith("logo") and f.endswith("png")]
    from metadrive.utils import get_np_random
    selected = get_np_random().choice(files)
    selected_file = AssetLoader.file_path("{}/{}".format(background_folder_name, selected))
    return selected_file


def get_logo_file():
    file = AssetLoader.file_path("logo-tiny.png")
    # assert os.path.exists(file)
    return file


def asset_version():
    from metadrive.version import asset_version
    return asset_version()



from metadrive.component.traffic_light.base_traffic_light import BaseTrafficLight
from nuplan.common.maps.maps_datatypes import TrafficLightStatusType


class NuplanTrafficLight(BaseTrafficLight):
    def set_status(self, status: TrafficLightStatusType):
        if status == TrafficLightStatusType.GREEN:
            self.set_green()
        elif status == TrafficLightStatusType.RED:
            self.set_red()
        elif status == TrafficLightStatusType.YELLOW:
            self.set_yellow()
        elif status == TrafficLightStatusType.UNKNOWN:
            self.set_unknown()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

import time

from panda3d.core import PNMImage, VirtualFileSystem, VirtualFileMountRamdisk
from panda3d.core import Shader

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject

__all__ = ("RPLoader", )


class timed_loading_operation(object):  # noqa # pylint: disable=invalid-name,too-few-public-methods
    """ Context manager for a synchronous loading operation, keeping track
    on how much time elapsed during the loading process, and warning about
    long loading times. """

    WARNING_COUNT = 0

    def __init__(self, resource):
        self.resource = resource
        if isinstance(self.resource, (list, tuple)):
            self.resource = ', '.join(self.resource)

    def __enter__(self):
        self.start_time = time.process_time()

    def __exit__(self, *args):
        duration = (time.process_time() - self.start_time) * 1000.0
        if duration > 80.0 and timed_loading_operation.WARNING_COUNT < 5:
            RPObject.global_debug("RPLoader", "Loading '" + self.resource + "' took", round(duration, 2), "ms")
            timed_loading_operation.WARNING_COUNT += 1
            if timed_loading_operation.WARNING_COUNT == 5:
                RPObject.global_debug("RPLoader", "Skipping further loading warnings (max warning count reached)")


class RPLoader(RPObject):
    """ Generic loader class used by the pipeline. All loading of assets happens
    here, which enables us to keep track of used resources """
    @classmethod
    def load_texture(cls, filename):
        """ Loads a 2D-texture from disk """
        with timed_loading_operation(filename):
            return Globals.base.loader.load_texture(filename)

    @classmethod
    def load_cube_map(cls, filename, read_mipmaps=False):
        """ Loads a cube map from disk """
        with timed_loading_operation(filename):
            return Globals.base.loader.load_cube_map(filename, readMipmaps=read_mipmaps)

    @classmethod
    def load_3d_texture(cls, filename):
        """ Loads a 3D-texture from disk """
        with timed_loading_operation(filename):
            return Globals.base.loader.load_3d_texture(filename)

    @classmethod
    def load_font(cls, filename):
        """ Loads a font from disk """
        with timed_loading_operation(filename):
            return Globals.base.loader.load_font(filename)

    @classmethod
    def load_shader(cls, *args):
        """ Loads a shader from disk """
        with timed_loading_operation(args):
            if len(args) == 1:
                return Shader.load_compute(Shader.SL_GLSL, args[0])
            return Shader.load(Shader.SL_GLSL, *args)

    @classmethod
    def load_model(cls, filename):
        """ Loads a model from disk """
        with timed_loading_operation(filename):
            return Globals.base.loader.load_model(filename)

    @classmethod
    def load_sliced_3d_texture(cls, fname, tile_size_x, tile_size_y=None, num_tiles=None):
        """ Loads a texture from the given filename and dimensions. If only
        one dimensions is specified, the other dimensions are assumed to be
        equal. This internally loads the texture into ram, splits it into smaller
        sub-images, and then calls the load_3d_texture from the Panda loader """

        tempfile_name = "/$$slice_loader_temp-" + str(time.time()) + "/"
        tile_size_y = tile_size_x if tile_size_y is None else tile_size_y
        num_tiles = tile_size_x if num_tiles is None else num_tiles

        # Load sliced image from disk
        tex_handle = cls.load_texture(fname)
        source = PNMImage()
        tex_handle.store(source)
        width = source.get_x_size()

        # Find slice properties
        num_cols = width // tile_size_x
        temp_img = PNMImage(tile_size_x, tile_size_y, source.get_num_channels(), source.get_maxval())

        # Construct a ramdisk to write the files to
        vfs = VirtualFileSystem.get_global_ptr()
        ramdisk = VirtualFileMountRamdisk()
        vfs.mount(ramdisk, tempfile_name, 0)

        # Extract all slices and write them to the virtual disk
        for z_slice in range(num_tiles):
            slice_x = (z_slice % num_cols) * tile_size_x
            slice_y = (z_slice // num_cols) * tile_size_y
            temp_img.copy_sub_image(source, 0, 0, slice_x, slice_y, tile_size_x, tile_size_y)
            temp_img.write(tempfile_name + str(z_slice) + ".png")

        # Load the de-sliced texture from the ramdisk
        texture_handle = cls.load_3d_texture(tempfile_name + "/#.png")
        vfs.unmount(ramdisk)

        return texture_handle



from metadrive.component.pgblock.curve import Curve
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.roundabout import Roundabout
from metadrive.component.pgblock.std_intersection import StdInterSection
from metadrive.component.pgblock.straight import Straight
from metadrive.component.pgblock.t_intersection import TInterSection
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.engine.asset_loader import initialize_asset_loader
from metadrive.tests.vis_block.vis_block_base import TestBlock
from metadrive.component.pg_space import Parameter

if __name__ == "__main__":
    FirstPGBlock.ENTRANCE_LENGTH = 0.5
    test = TestBlock(False)
    initialize_asset_loader(test)
    global_network = NodeRoadNetwork()
    blocks = []
    init_block = FirstPGBlock(global_network, 3.0, 3, test.render, test.world, 1)

    block = StdInterSection(1, init_block.get_socket(0), global_network, 1)
    block.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0
        }, test.render, test.world
    )

    block = Straight(2, block.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 388}, test.render, test.world)

    block = StdInterSection(3, block.get_socket(0), global_network, 1)
    block.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0
        }, test.render, test.world
    )

    block = Straight(4, block.get_socket(2), global_network, 1)
    block.construct_from_config({Parameter.length: 70}, test.render, test.world)

    t_1 = TInterSection(5, block.get_socket(0), global_network, 1)
    t_1.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(6, t_1.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 68}, test.render, test.world)

    t_last = TInterSection(100, block.get_socket(0), global_network, 1)
    t_last.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 2,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(101, t_last.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 9}, test.render, test.world)

    t_2 = TInterSection(7, block.get_socket(0), global_network, 1)
    t_2.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(8, t_2.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 70}, test.render, test.world)

    block = StdInterSection(9, block.get_socket(0), global_network, 1)
    block.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0
        }, test.render, test.world
    )

    block = Straight(10, block.get_socket(2), global_network, 1)
    block.construct_from_config({Parameter.length: 70}, test.render, test.world)

    t_3 = TInterSection(11, block.get_socket(0), global_network, 1)
    t_3.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(12, t_3.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 130}, test.render, test.world)

    t_4 = TInterSection(13, block.get_socket(0), global_network, 1)
    t_4.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(14, t_4.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 70}, test.render, test.world)

    block = StdInterSection(15, block.get_socket(0), global_network, 1)
    block.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0
        }, test.render, test.world
    )

    block = Straight(16, block.get_socket(2), global_network, 1)
    block.construct_from_config({Parameter.length: 259}, test.render, test.world)

    t_5 = TInterSection(17, block.get_socket(0), global_network, 1)
    t_5.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(18, t_5.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 47}, test.render, test.world)

    block = Straight(19, t_5.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 70}, test.render, test.world)

    t_6 = TInterSection(20, block.get_socket(0), global_network, 1)
    t_6.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(21, t_6.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 130}, test.render, test.world)

    t_7 = TInterSection(22, block.get_socket(0), global_network, 1)
    t_7.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 2
        }, test.render, test.world
    )

    block = Straight(23, t_7.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 50}, test.render, test.world)

    block = Straight(24, t_7.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 130}, test.render, test.world)

    t_8 = TInterSection(25, block.get_socket(0), global_network, 1)
    t_8.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 0,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(26, t_8.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 50}, test.render, test.world)

    t_9 = TInterSection(27, t_8.get_socket(0), global_network, 1)
    t_9.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 2
        }, test.render, test.world
    )

    o = Roundabout(28, t_9.get_socket(0), global_network, 1)
    o.construct_from_config(
        {
            Parameter.radius_exit: 5,
            Parameter.radius_inner: 30,
            Parameter.angle: 60
        }, test.render, test.world
    )

    curve = Curve(29, o.get_socket(0), global_network, 1)
    curve.construct_from_config(
        {
            Parameter.radius: 20,
            Parameter.dir: 0,
            Parameter.angle: 90,
            Parameter.length: 70
        }, test.render, test.world
    )

    block = Straight(30, t_6.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 120}, test.render, test.world)

    t_10 = TInterSection(31, block.get_socket(0), global_network, 1)
    t_10.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(32, t_10.get_socket(0), global_network, 1)
    block.construct_from_config({Parameter.length: 16}, test.render, test.world)

    t_11 = TInterSection(33, block.get_socket(0), global_network, 1)
    t_11.construct_from_config(
        {
            Parameter.radius: 10,
            Parameter.change_lane_num: 1,
            Parameter.decrease_increase: 0,
            Parameter.t_intersection_type: 0
        }, test.render, test.world
    )

    block = Straight(34, t_11.get_socket(1), global_network, 1)
    block.construct_from_config({Parameter.length: 40}, test.render, test.world)

    test.run()



"""

This script downloads real world sun position data and imports them into the
pipelines time of day system.

(c) 2016 tobpsr

"""

from __future__ import print_function, division

# You can get this data from http://sunpath.azurewebsites.net/
CONFIG = {
    "year": 2016,
    "month": 7,
    "day": 21,
    "latitude": 50.778228759765625,
    "longitude": 6.088640213012695,
    "height": 167,
    "precision": 6
}

if __name__ == "__main__":

    import sys
    import math

    sys.path.insert(0, "../../")
    from rplibs.six.moves import urllib  # pylint: disable=import-error

    CONFIG["end_day"] = CONFIG["day"]

    API = "http://www.nrel.gov/midc/apps/spa.pl?syear={year}&smonth={month}&sday={day}&eyear={year}&emonth={month}&eday={end_day}&step={precision}&stepunit=1&latitude={latitude}&longitude={longitude}&timezone=%2B1&elev={height}&press=1013&temp=20&dut1=0.0&deltat=64.797&azmrot=180&slope=0&refract=0.5667&field=2&field=35&field=36&field=38&zip=0"  # noqa

    url = API.format(**CONFIG)

    query_str = (
        "Are you sure you want to import the data? This overrides your "
        "current sun azimuth, altitude and intensity settings! (y/n): "
    )
    if sys.version_info.major >= 3:
        query_result = input(query_str)
    else:
        query_result = raw_input(query_str)

    if query_result.strip().lower() not in ["y", "yes", "j"]:
        print("Aborted.")
        sys.exit(0)

    print("Querying api ..")
    handle = urllib.request.urlopen(url)
    data = handle.read().decode("utf-8")
    handle.close()

    # Extract data
    print("Extracting data ..")
    lines = data.replace("\r", "").split("\n")[1:]

    # Load render pipeline api
    print("Loading plugin api ..")
    sys.path.insert(0, "../../")
    from rpcore.pluginbase.manager import PluginManager
    from rpcore.mount_manager import MountManager

    mount_mgr = MountManager(None)
    mount_mgr.mount()

    plugin_mgr = PluginManager(None)
    plugin_mgr.load()

    convert_to_linear = plugin_mgr.day_settings["scattering"]["sun_intensity"].get_linear_value

    hour, minutes = 0, 0

    data_points_azimuth = []
    data_points_altitude = []
    data_points_intensity = []

    for line in lines:
        if not line:
            break
        date, time, azim_angle, declination, ascension, elevation = line.split(",")

        float_azim = (float(azim_angle) + 180) / 360
        float_elev = (float(elevation) / 60) * 0.5 + 0.5

        float_time = (minutes + hour * 60) / (24 * 60)

        data_points_azimuth.append((float_time, float_azim))
        data_points_altitude.append((float_time, float_elev))

        # Approximated intensity
        approx_intensity = 2.0 * (1 - math.cos(math.pi * max(0, 8.0 + float(elevation)) / 60.0))
        approx_intensity = max(0, min(150.0, approx_intensity * 0.25))
        data_points_intensity.append((float_time, approx_intensity))

        minutes += CONFIG["precision"]
        if minutes >= 60:
            hour += 1
            minutes = minutes % 60

    print("Saving ..")
    settings = plugin_mgr.day_settings["scattering"]
    settings["sun_azimuth"].curves[0].control_points = data_points_azimuth
    settings["sun_altitude"].curves[0].control_points = data_points_altitude
    settings["sun_intensity"].curves[0].control_points = data_points_intensity

    plugin_mgr.save_daytime_overrides("/$$rpconfig/daytime.yaml")

    print("Done!")



from metadrive.component.sensors.rgb_camera import RGBCamera
from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "start_seed": 4,
            "stack_size": 5,
            # "debug": True,
            # "debug_panda3d": True,
            "vehicle_config": dict(image_source="main_camera"),
            "sensors": {
                "rgb_camera": (RGBCamera, 84, 84)
            },
            "interface_panel": ["dashboard", "rgb_camera"],
            "manual_control": False,
            "use_render": True,
            "image_observation": True,  # it is a switch telling metadrive to use rgb as observation
            "rgb_clip": True,  # clip rgb to range(0,1) instead of (0, 255)
            # "pstats": True,
        }
    )
    env.reset()
    # # print m to capture rgb observation
    env.engine.accept(
        "m", env.engine.get_sensor(env.vehicle.config["image_source"]).save_image, extraArgs=[env.vehicle]
    )
    import cv2

    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        assert env.observation_space.contains(o)
        # save
        rgb_cam = env.engine.get_sensor(env.vehicle.config["image_source"])
        rgb_cam.save_image(env.vehicle, name="{}.png".format(i))
        cv2.imshow('img', o["image"][..., -1])
        cv2.waitKey(1)

        # if env.config["use_render"]:
        # for i in range(ImageObservation.STACK_SIZE):
        #      ObservationType.show_gray_scale_array(o["image"][:, :, i])
        # image = env.render(mode="any str except human", text={"can you see me": i})
        # ObservationType.show_gray_scale_array(image)
        if tm or tc:
            # print("Reset")
            env.reset()
    env.close()



from metadrive.component.traffic_participants.pedestrian import Pedestrian
from metadrive.component.traffic_light.base_traffic_light import BaseTrafficLight
from metadrive.envs.metadrive_env import MetaDriveEnv


def vis_traffic_light(render=True, manual_control=False, debug=False):
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.,
            "traffic_mode": "hybrid",
            "manual_control": manual_control,
            "use_render": render,
            "debug": debug,
            "debug_static_world": debug,
            "map": "X",
            "window_size": (1200, 800),
            "show_coordinates": True,
            "vehicle_config": {
                "enable_reverse": True,
                "show_dest_mark": True
            },
        }
    )
    env.reset()
    try:
        # green
        env.reset()
        light = env.engine.spawn_object(
            BaseTrafficLight, lane=env.current_map.road_network.graph[">>>"]["1X1_0_"][0], pbr_model=True
        )
        light.set_green()
        test_success = False
        for s in range(1, 1000):
            if s < 100:
                light.set_green()
            elif s < 300:
                light.set_red()
            else:
                light.set_yellow()
            env.step([0, 1])
        light.destroy()

        # red test
        light = env.engine.spawn_object(
            BaseTrafficLight, lane=env.current_map.road_network.graph[">>>"]["1X1_0_"][0], pbr_model=True
        )
        light.set_red()
        test_success = False
        for s in range(1, 1000):
            env.step([0, 1])
        light.destroy()
        # yellow
        env.reset()
        light = env.engine.spawn_object(
            BaseTrafficLight, lane=env.current_map.road_network.graph[">>>"]["1X1_0_"][0], pbr_model=True
        )
        light.set_yellow()
        test_success = False
        for s in range(1, 1000):
            env.step([0, 1])
        light.destroy()

    finally:
        env.close()


if __name__ == "__main__":
    vis_traffic_light(True, manual_control=True)
